@@ -70,7 +70,8 @@ from bzrlib.send import (
 
 if getattr(sys, "frozen", None):
     # allow import additional libs from ./_lib for bzr.exe only
-    sys.path.append(os.path.normpath(os.path.join(os.path.dirname(__file__), '_lib')))
+    sys.path.append(os.path.normpath(
+        os.path.join(os.path.dirname(__file__), '_lib')))
 
 _versions_checked = False
 def lazy_check_versions():
@@ -85,7 +86,9 @@ def lazy_check_versions():
             "bzr-git: Please install dulwich, https://launchpad.net/dulwich")
     else:
         if dulwich_version < dulwich_minimum_version:
-            raise bzr_errors.DependencyNotPresent("dulwich", "bzr-git: Dulwich is too old; at least %d.%d.%d is required" % dulwich_minimum_version)
+            raise bzr_errors.DependencyNotPresent("dulwich",
+                "bzr-git: Dulwich is too old; at least %d.%d.%d is required" %
+                    dulwich_minimum_version)
 
 bzrdir.format_registry.register_lazy('git',
     "bzrlib.plugins.git.dir", "LocalGitBzrDirFormat",
@@ -109,6 +112,8 @@ class GitBzrDirFormat(bzrdir.BzrDirFormat):
 
     _lock_class = TransportLock
 
+    colocated_branches = True
+
     def is_supported(self):
         return True
 
@@ -128,33 +133,23 @@ class LocalGitBzrDirFormat(GitBzrDirFormat):
 
         """
         lazy_check_versions()
-        import dulwich
-        # we dont grok readonly - git isn't integrated with transport.
-        url = transport.base
-        if url.startswith('readonly+'):
-            url = url[len('readonly+'):]
-
-        try:
-            gitrepo = dulwich.repo.Repo(transport.local_abspath(".").encode(osutils._fs_enc))
-        except bzr_errors.NotLocalUrl:
-            raise bzr_errors.NotBranchError(path=transport.base)
+        from bzrlib.plugins.git.transportgit import TransportRepo
+        gitrepo = TransportRepo(transport)
         from bzrlib.plugins.git.dir import LocalGitDir, GitLockableFiles, GitLock
         lockfiles = GitLockableFiles(transport, GitLock())
         return LocalGitDir(transport, lockfiles, gitrepo, self)
 
     @classmethod
     def probe_transport(klass, transport):
-        """Our format is present if the transport ends in '.not/'."""
-        from bzrlib.transport.local import LocalTransport
-
-        if not isinstance(transport, LocalTransport):
+        try:
+            if not transport.has_any(['info/refs', '.git/branches',
+                                      'branches']):
+                raise bzr_errors.NotBranchError(path=transport.base)
+        except bzr_errors.NoSuchFile:
             raise bzr_errors.NotBranchError(path=transport.base)
-
-        # This should quickly filter out most things that are not
-        # git repositories, saving us the trouble from loading dulwich.
-        if not transport.has(".git") and not transport.has("objects"):
+        from bzrlib import urlutils
+        if urlutils.split(transport.base)[1] == ".git":
             raise bzr_errors.NotBranchError(path=transport.base)
-
         lazy_check_versions()
         import dulwich
         format = klass()
@@ -180,7 +175,7 @@ class LocalGitBzrDirFormat(GitBzrDirFormat):
                 "non-local transports")
         lazy_check_versions()
         from dulwich.repo import Repo
-        Repo.create(transport.local_abspath(".").encode(osutils._fs_enc))
+        Repo.init(transport.local_abspath(".").encode(osutils._fs_enc))
         return self.open(transport)
 
     def is_supported(self):
@@ -255,6 +250,8 @@ foreign_vcs_registry.register_lazy("git",
 plugin_cmds.register_lazy("cmd_git_import", [], "bzrlib.plugins.git.commands")
 plugin_cmds.register_lazy("cmd_git_object", ["git-objects", "git-cat"],
     "bzrlib.plugins.git.commands")
+plugin_cmds.register_lazy("cmd_git_refs", [], "bzrlib.plugins.git.commands")
+plugin_cmds.register_lazy("cmd_git_apply", [], "bzrlib.plugins.git.commands")
 
 def update_stanza(rev, stanza):
     mapping = getattr(rev, "mapping", None)
@@ -274,23 +271,28 @@ transport_server_registry.register_lazy('git',
     'Git Smart server protocol over TCP. (default port: 9418)')
 
 
-from bzrlib.repository import network_format_registry as repository_network_format_registry
+from bzrlib.repository import (
+    network_format_registry as repository_network_format_registry,
+    )
 repository_network_format_registry.register_lazy('git',
     'bzrlib.plugins.git.repository', 'GitRepositoryFormat')
 
-from bzrlib.bzrdir import network_format_registry as bzrdir_network_format_registry
+from bzrlib.bzrdir import (
+    network_format_registry as bzrdir_network_format_registry,
+    )
 bzrdir_network_format_registry.register('git', GitBzrDirFormat)
 
-
-def get_rich_root_format(stacked=False):
-    if stacked:
-        return bzrdir.format_registry.make_bzrdir("1.9-rich-root")
-    else:
-        return bzrdir.format_registry.make_bzrdir("default-rich-root")
-
 send_format_registry.register_lazy('git', 'bzrlib.plugins.git.send',
                                    'send_git', 'Git am-style diff format')
 
+try:
+    from bzrlib.diff import format_registry as diff_format_registry
+except ImportError:
+    pass
+else:
+    diff_format_registry.register_lazy('git', 'bzrlib.plugins.git.send',
+        'GitDiffTree', 'Git am-style diff format')
+
 def test_suite():
     from bzrlib.plugins.git import tests
     return tests.test_suite()
