@@ -1,5 +1,5 @@
 # Copyright (C) 2007 Canonical Ltd
-# Copyright (C) 2009 Jelmer Vernooij <jelmer@samba.org>
+# Copyright (C) 2009-2010 Jelmer Vernooij <jelmer@samba.org>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -35,14 +35,14 @@ from bzrlib import (
 from bzrlib.decorators import (
     needs_read_lock,
     )
+from bzrlib.revision import (
+    NULL_REVISION,
+    )
 from bzrlib.trace import (
     is_quiet,
     mutter,
     )
 
-from bzrlib.plugins.git import (
-    get_rich_root_format,
-    )
 from bzrlib.plugins.git.config import (
     GitBranchConfig,
     )
@@ -50,20 +50,17 @@ from bzrlib.plugins.git.errors import (
     NoPushSupport,
     NoSuchRef,
     )
+from bzrlib.plugins.git.refs import (
+    ref_to_branch_name,
+    extract_tags,
+    tag_name_to_ref,
+    )
 
 from bzrlib.foreign import ForeignBranch
 
 
-def extract_tags(refs):
-    ret = {}
-    for k,v in refs.iteritems():
-        if k.startswith("refs/tags/") and not k.endswith("^{}"):
-            v = refs.get(k+"^{}", v)
-            ret[k[len("refs/tags/"):]] = v
-    return ret
-
-
 class GitPullResult(branch.PullResult):
+    """Result of a pull from a Git branch."""
 
     def _lookup_revno(self, revid):
         assert isinstance(revid, str), "was %r" % revid
@@ -101,17 +98,27 @@ class LocalGitTagDict(tag.BasicTags):
                 mutter("Tag %s points at object %r that is not a commit, "
                        "ignoring", k, obj)
                 continue
-            ret[k] = self.branch.mapping.revision_id_foreign_to_bzr(v)
+            ret[k] = self.branch.lookup_foreign_revision_id(v)
         return ret
 
+    def _set_tag_dict(self, to_dict):
+        extra = set(self.repository._git.get_refs().keys())
+        for k, revid in to_dict.iteritems():
+            name = tag_name_to_ref(k)
+            if name in extra:
+                extra.remove(name)
+            self.set_tag(k, revid)
+        for name in extra:
+            if name.startswith("refs/tags/"):
+                del self.repository._git[name]
+
     def set_tag(self, name, revid):
-        self.repository._git.refs["refs/tags/%s" % name], _ = \
+        self.repository._git.refs[tag_name_to_ref(name)], _ = \
             self.branch.mapping.revision_id_bzr_to_foreign(revid)
 
 
 class DictTagDict(LocalGitTagDict):
 
-
     def __init__(self, branch, tags):
         super(DictTagDict, self).__init__(branch)
         self._tags = tags
@@ -120,7 +127,6 @@ class DictTagDict(LocalGitTagDict):
         return self._tags
 
 
-
 class GitBranchFormat(branch.BranchFormat):
 
     def get_format_description(self):
@@ -144,10 +150,22 @@ class GitBranchFormat(branch.BranchFormat):
             return LocalGitTagDict(branch)
 
 
+class GitReadLock(object):
+
+    def __init__(self, unlock):
+        self.unlock = unlock
+
+
+class GitWriteLock(object):
+
+    def __init__(self, unlock):
+        self.unlock = unlock
+
+
 class GitBranch(ForeignBranch):
     """An adapter to git repositories for bzr Branch objects."""
 
-    def __init__(self, bzrdir, repository, name, lockfiles, tagsdict=None):
+    def __init__(self, bzrdir, repository, ref, lockfiles, tagsdict=None):
         self.repository = repository
         self._format = GitBranchFormat()
         self.control_files = lockfiles
@@ -155,7 +173,8 @@ class GitBranch(ForeignBranch):
         super(GitBranch, self).__init__(repository.get_mapping())
         if tagsdict is not None:
             self.tags = DictTagDict(self, tagsdict)
-        self.name = name
+        self.ref = ref
+        self.name = ref_to_branch_name(ref)
         self._head = None
         self.base = bzrdir.root_transport.base
 
@@ -163,7 +182,7 @@ class GitBranch(ForeignBranch):
         """Return the most suitable metadir for a checkout of this branch.
         Weaves are used if this branch's repository uses weaves.
         """
-        return get_rich_root_format()
+        return bzrdir.format_registry.make_bzrdir("default")
 
     def get_child_submit_format(self):
         """Return the preferred format of submissions to this branch."""
@@ -177,7 +196,7 @@ class GitBranch(ForeignBranch):
 
         :return: Branch nick
         """
-        return self.name
+        return self.name or "HEAD"
 
     def _set_nick(self, nick):
         raise NotImplementedError
@@ -185,7 +204,8 @@ class GitBranch(ForeignBranch):
     nick = property(_get_nick, _set_nick)
 
     def __repr__(self):
-        return "%s(%r, %r)" % (self.__class__.__name__, self.repository.base, self.name)
+        return "<%s(%r, %r)>" % (self.__class__.__name__, self.repository.base,
+            self.ref or "HEAD")
 
     def generate_revision_history(self, revid, old_revid=None):
         # FIXME: Check that old_revid is in the ancestry of revid
@@ -194,6 +214,7 @@ class GitBranch(ForeignBranch):
 
     def lock_write(self):
         self.control_files.lock_write()
+        return GitWriteLock(self.unlock)
 
     def get_stacked_on_url(self):
         # Git doesn't do stacking (yet...)
@@ -210,6 +231,7 @@ class GitBranch(ForeignBranch):
 
     def lock_read(self):
         self.control_files.lock_read()
+        return GitReadLock(self.unlock)
 
     def is_locked(self):
         return self.control_files.is_locked()
@@ -225,16 +247,27 @@ class GitBranch(ForeignBranch):
         # perhaps should escape this ?
         if self.head is None:
             return revision.NULL_REVISION
-        return self.mapping.revision_id_foreign_to_bzr(self.head)
+        return self.lookup_foreign_revision_id(self.head)
 
     def _basic_push(self, target, overwrite=False, stop_revision=None):
         return branch.InterBranch.get(self, target)._basic_push(
             overwrite, stop_revision)
 
+    def lookup_foreign_revision_id(self, foreign_revid):
+        return self.repository.lookup_foreign_revision_id(foreign_revid,
+            self.mapping)
+
 
 class LocalGitBranch(GitBranch):
     """A local Git branch."""
 
+    def __init__(self, bzrdir, repository, name, lockfiles, tagsdict=None):
+        super(LocalGitBranch, self).__init__(bzrdir, repository, name,
+              lockfiles, tagsdict)
+        refs = repository._git.get_refs()
+        if not (name in refs.keys() or "HEAD" in refs.keys()):
+            raise errors.NotBranchError(self.base)
+
     def create_checkout(self, to_location, revision_id=None, lightweight=False,
         accelerator_tree=None, hardlink=False):
         if lightweight:
@@ -249,7 +282,7 @@ class LocalGitBranch(GitBranch):
             return tree
         else:
             return self._create_heavyweight_checkout(to_location, revision_id,
-            hardlink)
+                hardlink)
 
     def _create_heavyweight_checkout(self, to_location, revision_id=None,
                                      hardlink=False):
@@ -261,7 +294,7 @@ class LocalGitBranch(GitBranch):
         :return: WorkingTree object of checkout.
         """
         checkout_branch = bzrdir.BzrDir.create_branch_convenience(
-            to_location, force_new_tree=False, format=get_rich_root_format())
+            to_location, force_new_tree=False)
         checkout = checkout_branch.bzrdir
         checkout_branch.bind(self)
         # pull up to the specified revision_id to set the initial
@@ -279,7 +312,7 @@ class LocalGitBranch(GitBranch):
 
     def _get_head(self):
         try:
-            return self.repository._git.ref(self.name)
+            return self.repository._git.ref(self.ref or "HEAD")
         except KeyError:
             return None
 
@@ -287,13 +320,12 @@ class LocalGitBranch(GitBranch):
         self.set_last_revision(revid)
 
     def set_last_revision(self, revid):
-        (newhead, self.mapping) = self.mapping.revision_id_bzr_to_foreign(
-                revid)
+        (newhead, self.mapping) = self.repository.lookup_bzr_revision_id(revid)
         self.head = newhead
 
     def _set_head(self, value):
         self._head = value
-        self.repository._git.refs[self.name] = self._head
+        self.repository._git.refs[self.ref or "HEAD"] = self._head
         self._clear_cached_state()
 
     head = property(_get_head, _set_head)
@@ -317,15 +349,52 @@ class LocalGitBranch(GitBranch):
 
 class GitBranchPullResult(branch.PullResult):
 
+    def __init__(self):
+        super(GitBranchPullResult, self).__init__()
+        self.new_git_head = None
+        self._old_revno = None
+        self._new_revno = None
+
     def report(self, to_file):
         if not is_quiet():
             if self.old_revid == self.new_revid:
                 to_file.write('No revisions to pull.\n')
-            else:
+            elif self.new_git_head is not None:
                 to_file.write('Now on revision %d (git sha: %s).\n' %
                         (self.new_revno, self.new_git_head))
+            else:
+                to_file.write('Now on revision %d.\n' % (self.new_revno,))
         self._show_tag_conficts(to_file)
 
+    def _lookup_revno(self, revid):
+        assert isinstance(revid, str), "was %r" % revid
+        # Try in source branch first, it'll be faster
+        try:
+            return self.source_branch.revision_id_to_revno(revid)
+        except errors.NoSuchRevision:
+            # FIXME: Check using graph.find_distance_to_null() ?
+            return self.target_branch.revision_id_to_revno(revid)
+
+    def _get_old_revno(self):
+        if self._old_revno is not None:
+            return self._old_revno
+        return self._lookup_revno(self.old_revid)
+
+    def _set_old_revno(self, revno):
+        self._old_revno = revno
+
+    old_revno = property(_get_old_revno, _set_old_revno)
+
+    def _get_new_revno(self):
+        if self._new_revno is not None:
+            return self._new_revno
+        return self._lookup_revno(self.new_revid)
+
+    def _set_new_revno(self, revno):
+        self._new_revno = revno
+
+    new_revno = property(_get_new_revno, _set_new_revno)
+
 
 class GitBranchPushResult(branch.BranchPushResult):
 
@@ -361,36 +430,54 @@ class InterFromGitBranch(branch.GenericInterBranch):
                 not isinstance(target, GitBranch) and
                 (getattr(cls._get_interrepo(source, target), "fetch_objects", None) is not None))
 
-    def update_revisions(self, stop_revision=None, overwrite=False,
-        graph=None):
-        """See InterBranch.update_revisions()."""
+    def _update_revisions(self, stop_revision=None, overwrite=False,
+        graph=None, limit=None):
+        """Like InterBranch.update_revisions(), but with additions.
+
+        Compared to the `update_revisions()` below, this function takes a
+        `limit` argument that limits how many git commits will be converted
+        and returns the new git head.
+        """
         interrepo = self._get_interrepo(self.source, self.target)
-        self._head = None
-        self._last_revid = None
         def determine_wants(heads):
-            if not self.source.name in heads:
-                raise NoSuchRef(self.source.name, heads.keys())
+            if self.source.ref is not None and not self.source.ref in heads:
+                raise NoSuchRef(self.source.ref, heads.keys())
             if stop_revision is not None:
                 self._last_revid = stop_revision
-                self._head, mapping = self.source.repository.lookup_bzr_revision_id(
+                head, mapping = self.source.repository.lookup_bzr_revision_id(
                     stop_revision)
             else:
-                self._head = heads[self.source.name]
-                self._last_revid = \
-                    self.source.mapping.revision_id_foreign_to_bzr(self._head)
+                if self.source.ref is not None:
+                    head = heads[self.source.ref]
+                else:
+                    head = heads["HEAD"]
+                self._last_revid = self.source.lookup_foreign_revision_id(head)
             if self.target.repository.has_revision(self._last_revid):
                 return []
-            return [self._head]
-        interrepo.fetch_objects(determine_wants, self.source.mapping)
+            return [head]
+        pack_hint, head = interrepo.fetch_objects(
+            determine_wants, self.source.mapping, limit=limit)
+        if (pack_hint is not None and
+            self.target.repository._format.pack_compresses):
+            self.target.repository.pack(hint=pack_hint)
+        if head is not None:
+            self._last_revid = self.source.lookup_foreign_revision_id(head)
         if overwrite:
             prev_last_revid = None
         else:
             prev_last_revid = self.target.last_revision()
-        self.target.generate_revision_history(self._last_revid, prev_last_revid)
+        self.target.generate_revision_history(self._last_revid,
+            prev_last_revid)
+        return head
+
+    def update_revisions(self, stop_revision=None, overwrite=False,
+                         graph=None):
+        """See InterBranch.update_revisions()."""
+        self._update_revisions(stop_revision, overwrite, graph)
 
     def pull(self, overwrite=False, stop_revision=None,
              possible_transports=None, _hook_master=None, run_hooks=True,
-             _override_hook_target=None, local=False):
+             _override_hook_target=None, local=False, limit=None):
         """See Branch.pull.
 
         :param _hook_master: Private parameter - set the branch to
@@ -400,6 +487,8 @@ class InterFromGitBranch(branch.GenericInterBranch):
             so it should not run its hooks.
         :param _override_hook_target: Private parameter - set the branch to be
             supplied as the target_branch to pull hooks.
+        :param limit: Only import this many revisons.  `None`, the default,
+            means import all revisions.
         """
         # This type of branch can't be bound.
         if local:
@@ -415,14 +504,14 @@ class InterFromGitBranch(branch.GenericInterBranch):
             # We assume that during 'pull' the target repository is closer than
             # the source one.
             graph = self.target.repository.get_graph(self.source.repository)
-            result.old_revno, result.old_revid = \
+            (result.old_revno, result.old_revid) = \
                 self.target.last_revision_info()
-            self.update_revisions(stop_revision, overwrite=overwrite,
-                graph=graph)
-            result.new_git_head = self._head
+            result.new_git_head = self._update_revisions(
+                stop_revision, overwrite=overwrite, graph=graph, limit=limit)
             result.tag_conflicts = self.source.tags.merge_to(self.target.tags,
                 overwrite)
-            result.new_revno, result.new_revid = self.target.last_revision_info()
+            (result.new_revno, result.new_revid) = \
+                self.target.last_revision_info()
             if _hook_master:
                 result.master_branch = _hook_master
                 result.local_branch = result.target_branch
@@ -442,8 +531,8 @@ class InterFromGitBranch(branch.GenericInterBranch):
         result.target_branch = self.target
         graph = self.target.repository.get_graph(self.source.repository)
         result.old_revno, result.old_revid = self.target.last_revision_info()
-        self.update_revisions(stop_revision, overwrite=overwrite, graph=graph)
-        result.new_git_head = self._head
+        result.new_git_head = self._update_revisions(
+            stop_revision, overwrite=overwrite, graph=graph)
         result.tag_conflicts = self.source.tags.merge_to(self.target.tags,
             overwrite)
         result.new_revno, result.new_revid = self.target.last_revision_info()
@@ -464,6 +553,7 @@ class InterGitLocalRemoteBranch(InterGitBranch):
                 isinstance(target, RemoteGitBranch))
 
     def _basic_push(self, overwrite=False, stop_revision=None):
+        from dulwich.protocol import ZERO_SHA
         result = GitBranchPushResult()
         result.source_branch = self.source
         result.target_branch = self.target
@@ -471,14 +561,14 @@ class InterGitLocalRemoteBranch(InterGitBranch):
             stop_revision = self.source.last_revision()
         # FIXME: Check for diverged branches
         def get_changed_refs(old_refs):
-            result.old_revid = self.target.mapping.revision_id_foreign_to_bzr(old_refs.get("refs/heads/master", "0" * 40))
-            refs = { "refs/heads/master": self.source.repository.lookup_bzr_revision_id(stop_revision)[0] }
+            result.old_revid = self.target.lookup_foreign_revision_id(old_refs.get(self.target.ref, ZERO_SHA))
+            refs = { self.target.ref: self.source.repository.lookup_bzr_revision_id(stop_revision)[0] }
             result.new_revid = stop_revision
             for name, sha in self.source.repository._git.refs.as_dict("refs/tags").iteritems():
-                refs["refs/tags/%s" % name] = sha
+                refs[tag_name_to_ref(name)] = sha
             return refs
         self.target.repository.send_pack(get_changed_refs,
-                self.source.repository._git.object_store.generate_pack_contents)
+            self.source.repository._git.object_store.generate_pack_contents)
         return result
 
 
@@ -504,21 +594,21 @@ class InterGitRemoteLocalBranch(InterGitBranch):
 
     def update_tags(self, refs):
         for name, v in extract_tags(refs).iteritems():
-            revid = self.target.mapping.revision_id_foreign_to_bzr(v)
+            revid = self.target.lookup_foreign_revision_id(v)
             self.target.tags.set_tag(name, revid)
 
     def update_refs(self, stop_revision=None):
         interrepo = repository.InterRepository.get(self.source.repository,
             self.target.repository)
         if stop_revision is None:
-            refs = interrepo.fetch_refs(branches=["HEAD"])
-            stop_revision = self.target.mapping.revision_id_foreign_to_bzr(refs["HEAD"])
+            refs = interrepo.fetch(branches=["HEAD"])
+            stop_revision = self.target.lookup_foreign_revision_id(refs["HEAD"])
         else:
-            refs = interrepo.fetch_refs(revision_id=stop_revision)
+            refs = interrepo.fetch(revision_id=stop_revision)
         return refs, stop_revision
 
     def pull(self, stop_revision=None, overwrite=False,
-        possible_transports=None, local=False):
+        possible_transports=None, run_hooks=True,local=False):
         # This type of branch can't be bound.
         if local:
             raise errors.LocalRequiresBoundBranch()
@@ -533,12 +623,17 @@ class InterGitRemoteLocalBranch(InterGitBranch):
         return result
 
 
-class InterToGitBranch(branch.InterBranch):
+class InterToGitBranch(branch.GenericInterBranch):
     """InterBranch implementation that pulls from Git into bzr."""
 
+    def __init__(self, source, target):
+        super(InterToGitBranch, self).__init__(source, target)
+        self.interrepo = repository.InterRepository.get(source.repository,
+                                           target.repository)
+
     @staticmethod
     def _get_branch_formats_to_test():
-        return None, None
+        return []
 
     @classmethod
     def is_compatible(self, source, target):
@@ -548,33 +643,67 @@ class InterToGitBranch(branch.InterBranch):
     def update_revisions(self, *args, **kwargs):
         raise NoPushSupport()
 
-    def push(self, overwrite=True, stop_revision=None,
+    def _get_new_refs(self, stop_revision=None):
+        if stop_revision is None:
+            stop_revision = self.source.last_revision()
+        assert type(stop_revision) is str
+        main_ref = self.target.ref or "refs/heads/master"
+        refs = { main_ref: (None, stop_revision) }
+        for name, revid in self.source.tags.get_tag_dict().iteritems():
+            if self.source.repository.has_revision(revid):
+                refs[tag_name_to_ref(name)] = (None, revid)
+        return refs, main_ref
+
+    def pull(self, overwrite=False, stop_revision=None, local=False,
+             possible_transports=None):
+        from dulwich.protocol import ZERO_SHA
+        result = GitBranchPullResult()
+        result.source_branch = self.source
+        result.target_branch = self.target
+        new_refs, main_ref = self._get_new_refs(stop_revision)
+        def update_refs(old_refs):
+            refs = dict(old_refs)
+            # FIXME: Check for diverged branches
+            refs.update(new_refs)
+            return refs
+        old_refs, new_refs = self.interrepo.fetch_refs(update_refs)
+        result.old_revid = self.target.lookup_foreign_revision_id(
+            old_refs.get(main_ref, ZERO_SHA))
+        result.new_revid = new_refs[main_ref]
+        return result
+
+    def push(self, overwrite=False, stop_revision=None,
              _override_hook_source_branch=None):
-        raise NoPushSupport()
+        from dulwich.protocol import ZERO_SHA
+        result = GitBranchPushResult()
+        result.source_branch = self.source
+        result.target_branch = self.target
+        new_refs, main_ref = self._get_new_refs(stop_revision)
+        def update_refs(old_refs):
+            refs = dict(old_refs)
+            # FIXME: Check for diverged branches
+            refs.update(new_refs)
+            return refs
+        old_refs, new_refs = self.interrepo.fetch_refs(update_refs)
+        result.old_revid = self.target.lookup_foreign_revision_id(
+            old_refs.get(main_ref, ZERO_SHA))
+        result.new_revid = new_refs[main_ref]
+        return result
 
     def lossy_push(self, stop_revision=None):
         result = GitBranchPushResult()
         result.source_branch = self.source
         result.target_branch = self.target
-        try:
-            result.old_revid = self.target.last_revision()
-        except NoSuchRef:
-            result.old_revid = revision.NULL_REVISION
-        if stop_revision is None:
-            stop_revision = self.source.last_revision()
-        # FIXME: Check for diverged branches
-        refs = { "refs/heads/master": stop_revision }
-        for name, revid in self.source.tags.get_tag_dict().iteritems():
-            if self.source.repository.has_revision(revid):
-                refs["refs/tags/%s" % name] = revid
-        revidmap, new_refs = self.target.repository.dfetch_refs(
-            self.source.repository, refs)
-        if revidmap != {}:
-            self.target.generate_revision_history(revidmap[stop_revision])
-            result.new_revid = revidmap[stop_revision]
-        else:
-            result.new_revid = result.old_revid
-        result.revidmap = revidmap
+        new_refs, main_ref = self._get_new_refs(stop_revision)
+        def update_refs(old_refs):
+            refs = dict(old_refs)
+            # FIXME: Check for diverged branches
+            refs.update(new_refs)
+            return refs
+        result.revidmap, old_refs, new_refs = self.interrepo.dfetch_refs(
+            update_refs)
+        result.old_revid = old_refs.get(self.target.ref, (None, NULL_REVISION))[1]
+        result.new_revid = new_refs[main_ref][1]
         return result
 
 
