@@ -1,4 +1,4 @@
-# Copyright (C) 2009 Jelmer Vernooij <jelmer@samba.org>
+# Copyright (C) 2009-2010 Jelmer Vernooij <jelmer@samba.org>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -21,6 +21,7 @@
 from dulwich.index import (
     commit_tree,
     )
+import os
 import stat
 
 from bzrlib.repository import (
@@ -93,7 +94,8 @@ class GitCommitBuilder(CommitBuilder):
                 mode |= 0111
             self._any_changes = True
             self._blobs[path[1].encode("utf-8")] = (mode, sha)
-            yield file_id, path, (None, None)
+            file_sha1 = workingtree.get_file_sha1(file_id, path[1])
+            yield file_id, path[1], (file_sha1, os.lstat(workingtree.abspath(path[1])))
         # Fill in entries that were not changed
         basis_tree = workingtree.basis_tree()
         assert basis_tree.get_revision_id() == basis_revid
@@ -114,17 +116,22 @@ class GitCommitBuilder(CommitBuilder):
             if entry is None:
                 del self._blobs[path]
 
+    def _iterblobs(self):
+        return ((path, sha, mode) for (path, (mode, sha)) in self._blobs.iteritems())
+
     def commit(self, message):
         c = Commit()
         c.parents = [self.repository.lookup_bzr_revision_id(revid)[0] for revid in self.parents]
-        c.tree = commit_tree(self.store,
-                [(path, sha, mode) for (path, (mode, sha)) in self._blobs.iteritems()])
+        c.tree = commit_tree(self.store, self._iterblobs())
         c.committer = self._committer
         c.author = self._revprops.get('author', self._committer)
         c.commit_time = int(self._timestamp)
         c.author_time = int(self._timestamp)
         c.commit_timezone = self._timezone
         c.author_timezone = self._timezone
+        c.encoding = 'utf-8'
         c.message = message.encode("utf-8")
         self.store.add_object(c)
-        return self.repository.get_mapping().revision_id_foreign_to_bzr(c.id)
+        self._new_revision_id = self.repository.get_mapping().revision_id_foreign_to_bzr(c.id)
+        self.repository.commit_write_group()
+        return self._new_revision_id
