@@ -1,4 +1,5 @@
 # Copyright (C) 2007 Canonical Ltd
+# Copyright (C) 2010 Jelmer Vernooij
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -21,13 +22,13 @@ from bzrlib import (
     errors as bzr_errors,
     lockable_files,
     urlutils,
+    version_info as bzrlib_version,
     )
 
 LockWarner = getattr(lockable_files, "_LockWarner", None)
 
 from bzrlib.plugins.git import (
     LocalGitBzrDirFormat,
-    get_rich_root_format,
     )
 
 
@@ -75,7 +76,20 @@ class GitDir(bzrdir.BzrDir):
         return True
 
     def cloning_metadir(self, stacked=False):
-        return get_rich_root_format(stacked)
+        return bzrdir.format_registry.make_bzrdir("default")
+
+    def _branch_name_to_ref(self, name):
+        raise NotImplementedError(self._branch_name_to_ref)
+
+    if bzrlib_version >= (2, 2):
+        def open_branch(self, name=None, ignore_fallbacks=None,
+            unsupported=False):
+            return self._open_branch(name=name,
+                ignore_fallbacks=ignore_fallbacks, unsupported=unsupported)
+    else:
+        def open_branch(self, ignore_fallbacks=None, unsupported=False):
+            return self._open_branch(name=None,
+                ignore_fallbacks=ignore_fallbacks, unsupported=unsupported)
 
 
 class LocalGitDir(GitDir):
@@ -98,42 +112,88 @@ class LocalGitDir(GitDir):
         self._lockfiles = lockfiles
         self._mode_check_done = None
 
+    def _branch_name_to_ref(self, name):
+        from bzrlib.plugins.git.refs import branch_name_to_ref
+        ref = branch_name_to_ref(name, None)
+        if ref == "HEAD":
+            from dulwich.repo import SYMREF
+            refcontents = self._git.refs.read_ref(ref)
+            if refcontents.startswith(SYMREF):
+                ref = refcontents[len(SYMREF):]
+        return ref
+
     def is_control_filename(self, filename):
         return filename == '.git' or filename.startswith('.git/')
 
-    def get_branch_transport(self, branch_format):
+    def get_branch_transport(self, branch_format, name):
         if branch_format is None:
             return self.transport
         if isinstance(branch_format, LocalGitBzrDirFormat):
             return self.transport
         raise bzr_errors.IncompatibleFormat(branch_format, self._format)
 
-    get_repository_transport = get_branch_transport
-    get_workingtree_transport = get_branch_transport
+    def get_repository_transport(self, format):
+        if format is None:
+            return self.transport
+        if isinstance(format, LocalGitBzrDirFormat):
+            return self.transport
+        raise bzr_errors.IncompatibleFormat(format, self._format)
 
-    def open_branch(self, ignore_fallbacks=None):
+    def get_workingtree_transport(self, format):
+        if format is None:
+            return self.transport
+        if isinstance(format, LocalGitBzrDirFormat):
+            return self.transport
+        raise bzr_errors.IncompatibleFormat(format, self._format)
+
+    def _open_branch(self, name=None, ignore_fallbacks=None, unsupported=False):
         """'create' a branch for this dir."""
         repo = self.open_repository()
         from bzrlib.plugins.git.branch import LocalGitBranch
-        return LocalGitBranch(self, repo, "HEAD", self._lockfiles)
+        return LocalGitBranch(self, repo, self._branch_name_to_ref(name),
+            self._lockfiles)
+
+    def destroy_branch(self, name=None):
+        del self._git.refs[self._branch_name_to_ref(name)]
+
+    def list_branches(self):
+        ret = []
+        for name in self._git.get_refs():
+            if name.startswith("refs/heads/"):
+                ret.append(self.open_branch(name=name))
+        return ret
 
     def open_repository(self, shared=False):
         """'open' a repository for this dir."""
         return self._gitrepository_class(self, self._lockfiles)
 
     def open_workingtree(self, recommend_upgrade=True):
-        if not self._git.bare and self._git.has_index():
-            from bzrlib.plugins.git.workingtree import GitWorkingTree
-            return GitWorkingTree(self, self.open_repository(),
-                                                  self.open_branch())
+        if not self._git.bare:
+            from dulwich.errors import NoIndexPresent
+            repo = self.open_repository()
+            try:
+                index = repo._git.open_index()
+            except NoIndexPresent:
+                pass
+            else:
+                from bzrlib.plugins.git.workingtree import GitWorkingTree
+                try:
+                    branch = self.open_branch()
+                except bzr_errors.NotBranchError:
+                    pass
+                else:
+                    return GitWorkingTree(self, repo, branch, index)
         loc = urlutils.unescape_for_display(self.root_transport.base, 'ascii')
         raise bzr_errors.NoWorkingTree(loc)
 
     def create_repository(self, shared=False):
         return self.open_repository()
 
-    def create_branch(self):
-        return self.open_branch()
+    def create_branch(self, name=None):
+        refname = self._branch_name_to_ref(name)
+        from dulwich.protocol import ZERO_SHA
+        self._git.refs[refname or "HEAD"] = ZERO_SHA
+        return self.open_branch(name)
 
     def backup_bzrdir(self):
         if self._git.bare:
