@@ -1,4 +1,4 @@
-# Copyright (C) 2008 Jelmer Vernooij <jelmer@samba.org>
+# Copyright (C) 2008-2010 Jelmer Vernooij <jelmer@samba.org>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -17,21 +17,24 @@
 from dulwich.objects import (
     Commit,
     Tag,
+    Tree,
     S_ISGITLINK,
     )
 from dulwich.object_store import (
     tree_lookup_path,
     )
+from itertools import (
+    imap,
+    )
+import posixpath
 import re
 import stat
 
 from bzrlib import (
     debug,
-    lru_cache,
     osutils,
     trace,
     ui,
-    urlutils,
     )
 from bzrlib.errors import (
     BzrError,
@@ -50,22 +53,26 @@ from bzrlib.repository import (
 from bzrlib.revision import (
     NULL_REVISION,
     )
+from bzrlib.revisiontree import (
+    RevisionTree,
+    )
 from bzrlib.tsort import (
     topo_sort,
     )
 from bzrlib.versionedfile import (
-    FulltextContentFactory,
+    ChunkedContentFactory,
     )
 
 from bzrlib.plugins.git.mapping import (
     DEFAULT_FILE_MODE,
-    inventory_to_tree_and_blobs,
     mode_is_executable,
-    squash_revision,
+    mode_kind,
     warn_unusual_mode,
     )
 from bzrlib.plugins.git.object_store import (
     BazaarObjectStore,
+    LRUTreeCache,
+    _tree_to_objects,
     )
 from bzrlib.plugins.git.remote import (
     RemoteGitRepository,
@@ -77,11 +84,10 @@ from bzrlib.plugins.git.repository import (
     )
 
 
-MAX_INV_CACHE_SIZE = 50 * 1024 * 1024
-
-
-def import_git_blob(texts, mapping, path, hexsha, base_inv, base_ie, parent_id,
-    revision_id, parent_invs, shagitmap, lookup_object, executable, symlink):
+def import_git_blob(texts, mapping, path, name, (base_hexsha, hexsha), 
+        base_inv, parent_id, revision_id,
+        parent_invs, lookup_object, (base_mode, mode), store_updater,
+        lookup_file_id):
     """Import a git blob object into a bzr repository.
 
     :param texts: VersionedFiles to add to
@@ -89,28 +95,20 @@ def import_git_blob(texts, mapping, path, hexsha, base_inv, base_ie, parent_id,
     :param blob: A git blob
     :return: Inventory delta for this file
     """
-    file_id = mapping.generate_file_id(path)
-    if symlink:
+    if mapping.is_control_file(path):
+        return []
+    if base_hexsha == hexsha and base_mode == mode:
+        # If nothing has changed since the base revision, we're done
+        return []
+    file_id = lookup_file_id(path)
+    if stat.S_ISLNK(mode):
         cls = InventoryLink
     else:
         cls = InventoryFile
-    # We just have to hope this is indeed utf-8:
-    ie = cls(file_id, urlutils.basename(path).decode("utf-8"), parent_id)
-    ie.executable = executable
-    # See if this has changed at all
-    if base_ie is None:
-        base_sha = None
-    else:
-        try:
-            base_sha = shagitmap.lookup_blob(file_id, base_ie.revision)
-        except KeyError:
-            base_sha = None
-        else:
-            if (base_sha == hexsha and base_ie.executable == ie.executable
-                and base_ie.kind == ie.kind):
-                # If nothing has changed since the base revision, we're done
-                return [], []
-    if base_sha == hexsha and base_ie.kind == ie.kind:
+    ie = cls(file_id, name.decode("utf-8"), parent_id)
+    ie.executable = mode_is_executable(mode)
+    if base_hexsha == hexsha and mode_kind(base_mode) == mode_kind(mode):
+        base_ie = base_inv[base_inv.path2id(path)]
         ie.text_size = base_ie.text_size
         ie.text_sha1 = base_ie.text_sha1
         ie.symlink_target = base_ie.symlink_target
@@ -126,43 +124,47 @@ def import_git_blob(texts, mapping, path, hexsha, base_inv, base_ie, parent_id,
             ie.text_size = None
             ie.text_sha1 = None
         else:
-            ie.text_size = len(blob.data)
-            ie.text_sha1 = osutils.sha_string(blob.data)
+            ie.text_size = sum(imap(len, blob.chunked))
+            ie.text_sha1 = osutils.sha_strings(blob.chunked)
     # Check what revision we should store
     parent_keys = []
     for pinv in parent_invs:
-        if pinv.revision_id == base_inv.revision_id:
-            pie = base_ie
-            if pie is None:
-                continue
-        else:
-            try:
-                pie = pinv[file_id]
-            except NoSuchId:
-                continue
-        if pie.text_sha1 == ie.text_sha1 and pie.executable == ie.executable and pie.symlink_target == ie.symlink_target:
+        try:
+            pie = pinv[file_id]
+        except NoSuchId:
+            continue
+        if (pie.text_sha1 == ie.text_sha1 and
+            pie.executable == ie.executable and
+            pie.symlink_target == ie.symlink_target):
             # found a revision in one of the parents to use
             ie.revision = pie.revision
             break
-        parent_keys.append((file_id, pie.revision))
+        parent_key = (file_id, pie.revision)
+        if not parent_key in parent_keys:
+            parent_keys.append(parent_key)
     if ie.revision is None:
         # Need to store a new revision
         ie.revision = revision_id
-        assert file_id is not None
         assert ie.revision is not None
-        texts.insert_record_stream([FulltextContentFactory((file_id, ie.revision), tuple(parent_keys), ie.text_sha1, blob.data)])
-        shamap = [(hexsha, "blob", (ie.file_id, ie.revision))]
-    else:
-        shamap = []
+        if ie.kind == 'symlink':
+            chunks = []
+        else: 
+            chunks = blob.chunked
+        texts.insert_record_stream([
+            ChunkedContentFactory((file_id, ie.revision),
+                tuple(parent_keys), ie.text_sha1, chunks)])
     invdelta = []
-    if base_ie is not None:
-        old_path = base_inv.id2path(file_id)
-        if base_ie.kind == "directory":
-            invdelta.extend(remove_disappeared_children(old_path, base_ie.children, []))
+    if base_hexsha is not None:
+        old_path = path # Renames are not supported yet
+        if stat.S_ISDIR(base_mode):
+            invdelta.extend(remove_disappeared_children(base_inv, old_path,
+                lookup_object(base_hexsha), [], lookup_object))
     else:
         old_path = None
     invdelta.append((old_path, path, file_id, ie))
-    return (invdelta, shamap)
+    if base_hexsha != hexsha:
+        store_updater.add_object(blob, ie, path)
+    return invdelta
 
 
 class SubmodulesRequireSubtrees(BzrError):
@@ -170,39 +172,55 @@ class SubmodulesRequireSubtrees(BzrError):
     internal = False
 
 
-def import_git_submodule(texts, mapping, path, hexsha, base_inv, base_ie,
-    parent_id, revision_id, parent_invs, shagitmap, lookup_object):
-    file_id = mapping.generate_file_id(path)
-    ie = TreeReference(file_id, urlutils.basename(path.decode("utf-8")),
-        parent_id)
+def import_git_submodule(texts, mapping, path, name, (base_hexsha, hexsha),
+    base_inv, parent_id, revision_id, parent_invs, lookup_object,
+    (base_mode, mode), store_updater, lookup_file_id):
+    if base_hexsha == hexsha and base_mode == mode:
+        return [], {}
+    file_id = lookup_file_id(path)
+    ie = TreeReference(file_id, name.decode("utf-8"), parent_id)
     ie.revision = revision_id
-    if base_ie is None:
+    if base_hexsha is None:
         oldpath = None
     else:
         oldpath = path
-        if (base_ie.kind == ie.kind and
-            base_ie.reference_revision == ie.reference_revision):
-            ie.revision = base_ie.revision
     ie.reference_revision = mapping.revision_id_foreign_to_bzr(hexsha)
-    texts.insert_record_stream([FulltextContentFactory((file_id, ie.revision), (), None, "")])
+    texts.insert_record_stream([
+        ChunkedContentFactory((file_id, ie.revision), (), None, [])])
     invdelta = [(oldpath, path, file_id, ie)]
-    return invdelta, {}, {}
+    return invdelta, {}
 
 
-def remove_disappeared_children(path, base_children, existing_children):
+def remove_disappeared_children(base_inv, path, base_tree, existing_children,
+        lookup_object):
+    """Generate an inventory delta for removed children.
+
+    :param base_inv: Base inventory against which to generate the 
+        inventory delta.
+    :param path: Path to process
+    :param base_tree: Git Tree base object
+    :param existing_children: Children that still exist
+    :param lookup_object: Lookup a git object by its SHA1
+    :return: Inventory delta, as list
+    """
     ret = []
-    deletable = [(osutils.pathjoin(path, k), v) for k,v in base_children.iteritems() if k not in existing_children]
-    while deletable:
-        (path, ie) = deletable.pop()
-        ret.append((path, None, ie.file_id, None))
-        if ie.kind == "directory":
-            for name, child_ie in ie.children.iteritems():
-                deletable.append((osutils.pathjoin(path, name), child_ie))
+    for name, mode, hexsha in base_tree.iteritems():
+        if name in existing_children:
+            continue
+        c_path = posixpath.join(path, name.decode("utf-8"))
+        file_id = base_inv.path2id(c_path)
+        assert file_id is not None
+        ret.append((c_path, None, file_id, None))
+        if stat.S_ISDIR(mode):
+            ret.extend(remove_disappeared_children(
+                base_inv, c_path, lookup_object(hexsha), [], lookup_object))
     return ret
 
 
-def import_git_tree(texts, mapping, path, hexsha, base_inv, base_ie, parent_id,
-    revision_id, parent_invs, shagitmap, lookup_object, allow_submodules=False):
+def import_git_tree(texts, mapping, path, name, (base_hexsha, hexsha),
+        base_inv, parent_id, revision_id, parent_invs,
+        lookup_object, (base_mode, mode), store_updater,
+        lookup_file_id, allow_submodules=False):
     """Import a git tree object into a bzr repository.
 
     :param texts: VersionedFiles object to add to
@@ -211,110 +229,136 @@ def import_git_tree(texts, mapping, path, hexsha, base_inv, base_ie, parent_id,
     :param base_inv: Base inventory against which to return inventory delta
     :return: Inventory delta for this subtree
     """
+    if base_hexsha == hexsha and base_mode == mode:
+        # If nothing has changed since the base revision, we're done
+        return [], {}
     invdelta = []
-    file_id = mapping.generate_file_id(path)
+    file_id = lookup_file_id(path)
     # We just have to hope this is indeed utf-8:
-    ie = InventoryDirectory(file_id, urlutils.basename(path.decode("utf-8")),
-        parent_id)
-    if base_ie is None:
-        # Newly appeared here
+    ie = InventoryDirectory(file_id, name.decode("utf-8"), parent_id)
+    tree = lookup_object(hexsha)
+    if base_hexsha is None:
+        base_tree = None
+        old_path = None # Newly appeared here
+    else:
+        base_tree = lookup_object(base_hexsha)
+        old_path = path # Renames aren't supported yet
+    if base_tree is None or type(base_tree) is not Tree:
         ie.revision = revision_id
-        texts.insert_record_stream([FulltextContentFactory((file_id, ie.revision), (), None, "")])
-        invdelta.append((None, path, file_id, ie))
-    else:
-        # See if this has changed at all
-        try:
-            base_sha = shagitmap.lookup_tree(file_id, base_inv.revision_id)
-        except KeyError:
-            pass
-        else:
-            if base_sha == hexsha:
-                # If nothing has changed since the base revision, we're done
-                return [], {}, []
-        if base_ie.kind != "directory":
-            ie.revision = revision_id
-            texts.insert_record_stream([FulltextContentFactory((ie.file_id, ie.revision), (), None, "")])
-            invdelta.append((base_inv.id2path(ie.file_id), path, ie.file_id, ie))
-    if base_ie is not None and base_ie.kind == "directory":
-        base_children = base_ie.children
-    else:
-        base_children = {}
+        invdelta.append((old_path, path, ie.file_id, ie))
+        texts.insert_record_stream([
+            ChunkedContentFactory((ie.file_id, ie.revision), (), None, [])])
     # Remember for next time
     existing_children = set()
     child_modes = {}
-    shamap = []
-    tree = lookup_object(hexsha)
-    for mode, name, child_hexsha in tree.entries():
-        basename = name.decode("utf-8")
-        existing_children.add(basename)
-        child_path = osutils.pathjoin(path, name)
-        if stat.S_ISDIR(mode):
-            subinvdelta, grandchildmodes, subshamap = import_git_tree(
-                    texts, mapping, child_path, child_hexsha, base_inv,
-                    base_children.get(basename), file_id, revision_id,
-                    parent_invs, shagitmap, lookup_object,
-                    allow_submodules=allow_submodules)
-            invdelta.extend(subinvdelta)
-            child_modes.update(grandchildmodes)
-            shamap.extend(subshamap)
-        elif S_ISGITLINK(mode): # submodule
+    for child_mode, name, child_hexsha in tree.entries():
+        existing_children.add(name)
+        child_path = posixpath.join(path, name)
+        if type(base_tree) is Tree:
+            try:
+                child_base_mode, child_base_hexsha = base_tree[name]
+            except KeyError:
+                child_base_hexsha = None
+                child_base_mode = 0
+        else:
+            child_base_hexsha = None
+            child_base_mode = 0
+        if stat.S_ISDIR(child_mode):
+            subinvdelta, grandchildmodes = import_git_tree(texts, mapping,
+                child_path, name, (child_base_hexsha, child_hexsha), base_inv,
+                file_id, revision_id, parent_invs, lookup_object, 
+                (child_base_mode, child_mode), store_updater, lookup_file_id,
+                allow_submodules=allow_submodules)
+        elif S_ISGITLINK(child_mode): # submodule
             if not allow_submodules:
                 raise SubmodulesRequireSubtrees()
-            subinvdelta, grandchildmodes, subshamap = import_git_submodule(
-                    texts, mapping, child_path, child_hexsha, base_inv, base_children.get(basename),
-                    file_id, revision_id, parent_invs, shagitmap, lookup_object)
-            invdelta.extend(subinvdelta)
-            child_modes.update(grandchildmodes)
-            shamap.extend(subshamap)
+            subinvdelta, grandchildmodes = import_git_submodule(texts, mapping,
+                child_path, name, (child_base_hexsha, child_hexsha), base_inv,
+                file_id, revision_id, parent_invs, lookup_object,
+                (child_base_mode, child_mode), store_updater, lookup_file_id)
         else:
-            subinvdelta, subshamap = import_git_blob(texts, mapping,
-                    child_path, child_hexsha, base_inv, base_children.get(basename), file_id,
-                    revision_id, parent_invs, shagitmap, lookup_object,
-                    mode_is_executable(mode), stat.S_ISLNK(mode))
-            invdelta.extend(subinvdelta)
-            shamap.extend(subshamap)
-        if mode not in (stat.S_IFDIR, DEFAULT_FILE_MODE,
+            subinvdelta = import_git_blob(texts, mapping, child_path, name,
+                (child_base_hexsha, child_hexsha), base_inv, file_id,
+                revision_id, parent_invs, lookup_object,
+                (child_base_mode, child_mode), store_updater, lookup_file_id)
+            grandchildmodes = {}
+        child_modes.update(grandchildmodes)
+        invdelta.extend(subinvdelta)
+        if child_mode not in (stat.S_IFDIR, DEFAULT_FILE_MODE,
                         stat.S_IFLNK, DEFAULT_FILE_MODE|0111):
-            child_modes[child_path] = mode
+            child_modes[child_path] = child_mode
     # Remove any children that have disappeared
-    if base_ie is not None and base_ie.kind == "directory":
-        invdelta.extend(remove_disappeared_children(base_inv.id2path(file_id),
-            base_children, existing_children))
-    shamap.append((hexsha, "tree", (file_id, revision_id)))
-    return invdelta, child_modes, shamap
+    if base_tree is not None and type(base_tree) is Tree:
+        invdelta.extend(remove_disappeared_children(base_inv, old_path, 
+            base_tree, existing_children, lookup_object))
+    store_updater.add_object(tree, ie, path)
+    return invdelta, child_modes
 
 
-def approx_inv_size(inv):
-    # Very rough estimate, 1k per inventory entry
-    return len(inv) * 1024
+def verify_commit_reconstruction(target_git_object_retriever, lookup_object,
+    o, rev, ret_tree, parent_trees, mapping, unusual_modes):
+    new_unusual_modes = mapping.export_unusual_file_modes(rev)
+    if new_unusual_modes != unusual_modes:
+        raise AssertionError("unusual modes don't match: %r != %r" % (
+            unusual_modes, new_unusual_modes))
+    # Verify that we can reconstruct the commit properly
+    rec_o = target_git_object_retriever._reconstruct_commit(rev, o.tree, True)
+    if rec_o != o:
+        raise AssertionError("Reconstructed commit differs: %r != %r" % (
+            rec_o, o))
+    diff = []
+    new_objs = {}
+    for path, obj, ie in _tree_to_objects(ret_tree, parent_trees,
+        target_git_object_retriever._cache.idmap, unusual_modes, mapping.BZR_DUMMY_FILE):
+        old_obj_id = tree_lookup_path(lookup_object, o.tree, path)[1]
+        new_objs[path] = obj
+        if obj.id != old_obj_id:
+            diff.append((path, lookup_object(old_obj_id), obj))
+    for (path, old_obj, new_obj) in diff:
+        while (old_obj.type_name == "tree" and
+               new_obj.type_name == "tree" and
+               sorted(old_obj) == sorted(new_obj)):
+            for name in old_obj:
+                if old_obj[name][0] != new_obj[name][0]:
+                    raise AssertionError("Modes for %s differ: %o != %o" %
+                        (path, old_obj[name][0], new_obj[name][0]))
+                if old_obj[name][1] != new_obj[name][1]:
+                    # Found a differing child, delve deeper
+                    path = posixpath.join(path, name)
+                    old_obj = lookup_object(old_obj[name][1])
+                    new_obj = new_objs[path]
+                    break
+        raise AssertionError("objects differ for %s: %r != %r" % (path,
+            old_obj, new_obj))
 
 
 def import_git_commit(repo, mapping, head, lookup_object,
-                      target_git_object_retriever, parent_invs_cache):
+                      target_git_object_retriever, trees_cache):
     o = lookup_object(head)
-    rev = mapping.import_commit(o)
+    rev = mapping.import_commit(o,
+            lambda x: target_git_object_retriever.lookup_git_sha(x)[1][0])
     # We have to do this here, since we have to walk the tree and
     # we need to make sure to import the blobs / trees with the right
     # path; this may involve adding them more than once.
-    parent_invs = []
-    for parent_id in rev.parent_ids:
-        try:
-            parent_invs.append(parent_invs_cache[parent_id])
-        except KeyError:
-            parent_inv = repo.get_inventory(parent_id)
-            parent_invs.append(parent_inv)
-            parent_invs_cache[parent_id] = parent_inv
-    if parent_invs == []:
+    parent_trees = trees_cache.revision_trees(rev.parent_ids)
+    if parent_trees == []:
         base_inv = Inventory(root_id=None)
-        base_ie = None
+        base_tree = None
+        base_mode = None
     else:
-        base_inv = parent_invs[0]
-        base_ie = base_inv.root
-    inv_delta, unusual_modes, shamap = import_git_tree(repo.texts,
-            mapping, "", o.tree, base_inv, base_ie, None, rev.revision_id,
-            parent_invs, target_git_object_retriever._idmap, lookup_object,
+        base_inv = parent_trees[0].inventory
+        base_tree = lookup_object(o.parents[0]).tree
+        base_mode = stat.S_IFDIR
+    store_updater = target_git_object_retriever._get_updater(rev)
+    store_updater.add_object(o, None, None)
+    lookup_file_id = mapping.get_fileid_map(lookup_object, o.tree).lookup_file_id
+    inv_delta, unusual_modes = import_git_tree(repo.texts,
+            mapping, "", u"", (base_tree, o.tree), base_inv, 
+            None, rev.revision_id, [p.inventory for p in parent_trees],
+            lookup_object, (base_mode, stat.S_IFDIR), store_updater,
+            lookup_file_id,
             allow_submodules=getattr(repo._format, "supports_tree_reference", False))
-    target_git_object_retriever._idmap.add_entries(shamap)
+    store_updater.finish()
     if unusual_modes != {}:
         for path, mode in unusual_modes.iteritems():
             warn_unusual_mode(rev.foreign_revid, path, mode)
@@ -327,40 +371,33 @@ def import_git_commit(repo, mapping, head, lookup_object,
     rev.inventory_sha1, inv = repo.add_inventory_by_delta(basis_id,
               inv_delta, rev.revision_id, rev.parent_ids,
               base_inv)
-    parent_invs_cache[rev.revision_id] = inv
+    ret_tree = RevisionTree(repo, inv, rev.revision_id)
+    trees_cache.add(ret_tree)
     repo.add_revision(rev.revision_id, rev)
     if "verify" in debug.debug_flags:
-        new_unusual_modes = mapping.export_unusual_file_modes(rev)
-        if new_unusual_modes != unusual_modes:
-            raise AssertionError("unusual modes don't match: %r != %r" % (unusual_modes, new_unusual_modes))
-        objs = inventory_to_tree_and_blobs(inv, repo.texts, mapping, unusual_modes)
-        for sha1, newobj, path in objs:
-            assert path is not None
-            oldobj = tree_lookup_path(lookup_object, o.tree, path)
-            if oldobj != newobj:
-                raise AssertionError("%r != %r in %s" % (oldobj, newobj, path))
+        verify_commit_reconstruction(target_git_object_retriever, 
+            lookup_object, o, rev, ret_tree, parent_trees, mapping,
+            unusual_modes)
 
 
-def import_git_objects(repo, mapping, object_iter, target_git_object_retriever,
-        heads, pb=None):
+def import_git_objects(repo, mapping, object_iter,
+    target_git_object_retriever, heads, pb=None, limit=None):
     """Import a set of git objects into a bzr repository.
 
     :param repo: Target Bazaar repository
     :param mapping: Mapping to use
     :param object_iter: Iterator over Git objects.
+    :return: Tuple with pack hints and last imported revision id
     """
-    target_git_object_retriever._idmap.start_write_group() # FIXME: try/finally
     def lookup_object(sha):
         try:
             return object_iter[sha]
         except KeyError:
             return target_git_object_retriever[sha]
-    # TODO: a more (memory-)efficient implementation of this
     graph = []
     checked = set()
-    heads = list(heads)
-    parent_invs_cache = lru_cache.LRUSizeCache(compute_size=approx_inv_size,
-                                               max_size=MAX_INV_CACHE_SIZE)
+    heads = list(set(heads))
+    trees_cache = LRUTreeCache(repo)
     # Find and convert commit objects
     while heads:
         if pb is not None:
@@ -370,46 +407,54 @@ def import_git_objects(repo, mapping, object_iter, target_git_object_retriever,
         try:
             o = lookup_object(head)
         except KeyError:
-            trace.mutter('missing head %s', head)
             continue
         if isinstance(o, Commit):
-            rev = mapping.import_commit(o)
+            rev = mapping.import_commit(o, lambda x: None)
             if repo.has_revision(rev.revision_id):
                 continue
-            squash_revision(repo, rev)
             graph.append((o.id, o.parents))
-            target_git_object_retriever._idmap.add_entry(o.id, "commit",
-                    (rev.revision_id, o.tree))
             heads.extend([p for p in o.parents if p not in checked])
         elif isinstance(o, Tag):
-            heads.append(o.object[1])
+            if o.object[1] not in checked:
+                heads.append(o.object[1])
         else:
             trace.warning("Unable to import head object %r" % o)
         checked.add(o.id)
     del checked
     # Order the revisions
     # Create the inventory objects
-    batch_size = 100
+    batch_size = 1000
     revision_ids = topo_sort(graph)
     pack_hints = []
+    if limit is not None:
+        revision_ids = revision_ids[:limit]
+    last_imported = None
     for offset in range(0, len(revision_ids), batch_size):
-        repo.start_write_group()
+        target_git_object_retriever.start_write_group() 
         try:
-            for i, head in enumerate(revision_ids[offset:offset+batch_size]):
-                if pb is not None:
-                    pb.update("fetching revisions", offset+i, len(revision_ids))
-                import_git_commit(repo, mapping, head, lookup_object,
-                                  target_git_object_retriever,
-                                  parent_invs_cache)
+            repo.start_write_group()
+            try:
+                for i, head in enumerate(
+                    revision_ids[offset:offset+batch_size]):
+                    if pb is not None:
+                        pb.update("fetching revisions", offset+i,
+                                  len(revision_ids))
+                    import_git_commit(repo, mapping, head, lookup_object,
+                        target_git_object_retriever, trees_cache)
+                    last_imported = head
+            except:
+                repo.abort_write_group()
+                raise
+            else:
+                hint = repo.commit_write_group()
+                if hint is not None:
+                    pack_hints.extend(hint)
         except:
-            repo.abort_write_group()
+            target_git_object_retriever.abort_write_group()
             raise
         else:
-            hint = repo.commit_write_group()
-            if hint is not None:
-                pack_hints.extend(hint)
-    target_git_object_retriever._idmap.commit_write_group()
-    return pack_hints
+            target_git_object_retriever.commit_write_group()
+    return pack_hints, last_imported
 
 
 class InterGitRepository(InterRepository):
@@ -424,17 +469,12 @@ class InterGitRepository(InterRepository):
         """See InterRepository.copy_content."""
         self.fetch(revision_id, pb, find_ghosts=False)
 
-    def fetch(self, revision_id=None, pb=None, find_ghosts=False, mapping=None,
-            fetch_spec=None):
-        self.fetch_refs(revision_id=revision_id, pb=pb, find_ghosts=find_ghosts,
-                mapping=mapping, fetch_spec=fetch_spec)
-
 
 class InterGitNonGitRepository(InterGitRepository):
     """Base InterRepository that copies revisions from a Git into a non-Git
     repository."""
 
-    def fetch_refs(self, revision_id=None, pb=None, find_ghosts=False,
+    def fetch(self, revision_id=None, pb=None, find_ghosts=False,
               mapping=None, fetch_spec=None):
         if mapping is None:
             mapping = self.source.get_mapping()
@@ -450,16 +490,11 @@ class InterGitNonGitRepository(InterGitRepository):
             if interesting_heads is None:
                 ret = [sha for (ref, sha) in refs.iteritems() if not ref.endswith("^{}")]
             else:
-                ret = [mapping.revision_id_bzr_to_foreign(revid)[0] for revid in interesting_heads if revid not in (None, NULL_REVISION)]
-            return [rev for rev in ret if not self.target.has_revision(mapping.revision_id_foreign_to_bzr(rev))]
-        pack_hint = self.fetch_objects(determine_wants, mapping, pb)
+                ret = [self.source.lookup_bzr_revision_id(revid)[0] for revid in interesting_heads if revid not in (None, NULL_REVISION)]
+            return [rev for rev in ret if not self.target.has_revision(self.source.lookup_foreign_revision_id(rev))]
+        (pack_hint, _) = self.fetch_objects(determine_wants, mapping, pb)
         if pack_hint is not None and self.target._format.pack_compresses:
             self.target.pack(hint=pack_hint)
-        if interesting_heads is not None:
-            present_interesting_heads = self.target.has_revisions(interesting_heads)
-            missing_interesting_heads = set(interesting_heads) - present_interesting_heads
-            if missing_interesting_heads:
-                raise AssertionError("Missing interesting heads: %r" % missing_interesting_heads)
         return self._refs
 
 
@@ -486,7 +521,7 @@ class InterRemoteGitNonGitRepository(InterGitNonGitRepository):
         map(all_parents.update, parent_map.itervalues())
         return set(all_revs) - all_parents
 
-    def fetch_objects(self, determine_wants, mapping, pb=None):
+    def fetch_objects(self, determine_wants, mapping, pb=None, limit=None):
         def progress(text):
             report_git_progress(pb, text)
         store = BazaarObjectStore(self.target, mapping)
@@ -506,11 +541,10 @@ class InterRemoteGitNonGitRepository(InterGitNonGitRepository):
             if pb is None:
                 create_pb = pb = ui.ui_factory.nested_progress_bar()
             try:
-                objects_iter = self.source.fetch_objects(
-                            record_determine_wants, graph_walker,
-                            store.get_raw, progress)
+                objects_iter = self.source.fetch_objects(record_determine_wants,
+                    graph_walker, store.get_raw, progress)
                 return import_git_objects(self.target, mapping,
-                    objects_iter, store, recorded_wants, pb)
+                    objects_iter, store, recorded_wants, pb, limit)
             finally:
                 if create_pb:
                     create_pb.finished()
@@ -520,17 +554,19 @@ class InterRemoteGitNonGitRepository(InterGitNonGitRepository):
     @staticmethod
     def is_compatible(source, target):
         """Be compatible with GitRepository."""
-        # FIXME: Also check target uses VersionedFile
         return (isinstance(source, RemoteGitRepository) and
                 target.supports_rich_root() and
-                not isinstance(target, GitRepository))
+                not isinstance(target, GitRepository) and
+                target.texts is not None)
 
 
 class InterLocalGitNonGitRepository(InterGitNonGitRepository):
     """InterRepository that copies revisions from a local Git into a non-Git
     repository."""
 
-    def fetch_objects(self, determine_wants, mapping, pb=None):
+    def fetch_objects(self, determine_wants, mapping, pb=None, limit=None):
+        """Fetch objects.
+        """
         wants = determine_wants(self.source._git.get_refs())
         create_pb = None
         if pb is None:
@@ -540,8 +576,8 @@ class InterLocalGitNonGitRepository(InterGitNonGitRepository):
             self.target.lock_write()
             try:
                 return import_git_objects(self.target, mapping,
-                    self.source._git.object_store, target_git_object_retriever,
-                    wants, pb)
+                    self.source._git.object_store,
+                    target_git_object_retriever, wants, pb, limit)
             finally:
                 self.target.unlock()
         finally:
@@ -551,10 +587,10 @@ class InterLocalGitNonGitRepository(InterGitNonGitRepository):
     @staticmethod
     def is_compatible(source, target):
         """Be compatible with GitRepository."""
-        # FIXME: Also check target uses VersionedFile
         return (isinstance(source, LocalGitRepository) and
                 target.supports_rich_root() and
-                not isinstance(target, GitRepository))
+                not isinstance(target, GitRepository) and
+                target.texts is not None)
 
 
 class InterGitGitRepository(InterGitRepository):
@@ -585,7 +621,7 @@ class InterGitGitRepository(InterGitRepository):
         else:
             raise AssertionError
 
-    def fetch_refs(self, revision_id=None, pb=None, find_ghosts=False,
+    def fetch(self, revision_id=None, pb=None, find_ghosts=False,
               mapping=None, fetch_spec=None, branches=None):
         if mapping is None:
             mapping = self.source.get_mapping()
@@ -600,7 +636,7 @@ class InterGitGitRepository(InterGitRepository):
             determine_wants = r.object_store.determine_wants_all
         else:
             determine_wants = lambda x: [y for y in args if not y in r.object_store]
-        return self.fetch_objects(determine_wants, mapping)
+        return self.fetch_objects(determine_wants, mapping)[0]
 
 
     @staticmethod
