@@ -50,7 +50,8 @@ class GitInventoryEntry(inventory.InventoryEntry):
         self.path = path
         self.revision = self._inventory.revision_id
         self.executable = executable
-        self.file_id = self._inventory.mapping.generate_file_id(path.encode('utf-8'))
+        self.file_id = self._inventory.fileid_map.lookup_file_id(
+            path.encode('utf-8'))
 
     @property
     def object(self):
@@ -75,7 +76,7 @@ class GitInventoryFile(GitInventoryEntry):
 
     @property
     def text_sha1(self):
-        return osutils.sha_string(self.object.data)
+        return osutils.sha_strings(self.object.chunked)
 
     @property
     def text_size(self):
@@ -126,6 +127,7 @@ class GitInventoryLink(GitInventoryEntry):
 
     def copy(self):
         other = inventory.InventoryLink(self.file_id, self.name, self.parent_id)
+        other.executable = self.executable
         other.symlink_target = self.symlink_target
         other.revision = self.revision
         return other
@@ -177,12 +179,15 @@ class GitInventoryDirectory(GitInventoryEntry):
         for mode, name, hexsha in self.object.entries():
             basename = name.decode("utf-8")
             child_path = osutils.pathjoin(self.path, basename)
+            if self._inventory.mapping.is_control_file(child_path):
+                continue
             executable = mode_is_executable(mode)
             kind_class = {'directory': GitInventoryDirectory,
                           'file': GitInventoryFile,
                           'symlink': GitInventoryLink,
                           'tree-reference': GitInventoryTreeReference}[mode_kind(mode)]
-            self._children[basename] = kind_class(self._inventory, self.file_id, hexsha, child_path, basename, executable)
+            self._children[basename] = kind_class(self._inventory,
+                self.file_id, hexsha, child_path, basename, executable)
 
     def copy(self):
         other = inventory.InventoryDirectory(self.file_id, self.name,
@@ -195,9 +200,10 @@ class GitInventoryDirectory(GitInventoryEntry):
 
 class GitInventory(inventory.Inventory):
 
-    def __init__(self, tree_id, mapping, store, revision_id):
+    def __init__(self, tree_id, mapping, fileid_map, store, revision_id):
         super(GitInventory, self).__init__(revision_id=revision_id)
         self.store = store
+        self.fileid_map = fileid_map
         self.mapping = mapping
         self.root = GitInventoryDirectory(self, None, tree_id, u"", u"", False)
 
@@ -225,7 +231,7 @@ class GitInventory(inventory.Inventory):
             return False
 
     def id2path(self, file_id):
-        path = self.mapping.parse_file_id(file_id)
+        path = self.fileid_map.lookup_path(file_id)
         try:
             ie = self._get_ie(path)
             assert ie.path == path
@@ -241,7 +247,7 @@ class GitInventory(inventory.Inventory):
     def __getitem__(self, file_id):
         if file_id == inventory.ROOT_ID:
             return self.root
-        path = self.mapping.parse_file_id(file_id)
+        path = self.fileid_map.lookup_path(file_id)
         try:
             return self._get_ie(path)
         except KeyError:
@@ -251,20 +257,24 @@ class GitInventory(inventory.Inventory):
 class GitIndexInventory(inventory.Inventory):
     """Inventory that retrieves its contents from an index file."""
 
-    def __init__(self, basis_inventory, mapping, index, store):
+    def __init__(self, basis_inventory, fileid_map, index, store):
         super(GitIndexInventory, self).__init__(revision_id=None, root_id=basis_inventory.root.file_id)
         self.basis_inv = basis_inventory
-        self.mapping = mapping
+        self.fileid_map = fileid_map
         self.index = index
         self._contents_read = False
+        self.store = store
         self.root = self.add_path("", 'directory',
-            self.mapping.generate_file_id(""), None)
+            self.fileid_map.lookup_file_id(""), None)
 
     def iter_entries_by_dir(self, specific_file_ids=None, yield_parents=False):
         self._read_contents()
-        return super(GitIndexInventory, self).iter_entries_by_dir(specific_file_ids=specific_file_ids, yield_parents=yield_parents)
+        return super(GitIndexInventory, self).iter_entries_by_dir(
+            specific_file_ids=specific_file_ids, yield_parents=yield_parents)
 
     def has_id(self, file_id):
+        if type(file_id) != str:
+            raise AssertionError
         try:
             self.id2path(file_id)
             return True
@@ -278,7 +288,9 @@ class GitIndexInventory(inventory.Inventory):
         return super(GitIndexInventory, self).has_filename(path)
 
     def id2path(self, file_id):
-        path = self.mapping.parse_file_id(file_id)
+        if type(file_id) != str:
+            raise AssertionError
+        path = self.fileid_map.lookup_path(file_id)
         if path in self.index:
             return path
         self._read_contents()
@@ -286,9 +298,13 @@ class GitIndexInventory(inventory.Inventory):
 
     def path2id(self, path):
         if path in self.index:
-            return self.mapping.generate_file_id(path)
-        self._read_contents()
-        return super(GitIndexInventory, self).path2id(path)
+            file_id = self.fileid_map.lookup_file_id(path)
+        else:
+            self._read_contents()
+            file_id = super(GitIndexInventory, self).path2id(path)
+        if type(file_id) is not str:
+            raise AssertionError
+        return file_id
 
     def __getitem__(self, file_id):
         self._read_contents()
@@ -311,9 +327,11 @@ class GitIndexInventory(inventory.Inventory):
                 except KeyError:
                     old_ie = None
                 if old_ie is None:
-                    file_id = self.mapping.generate_file_id(path)
+                    file_id = self.fileid_map.lookup_file_id(path)
                 else:
                     file_id = old_ie.file_id
+                if type(file_id) != str:
+                    raise AssertionError
                 kind = mode_kind(mode)
                 if old_ie is not None and old_ie.hexsha == sha:
                     # Hasn't changed since basis inv
@@ -322,7 +340,7 @@ class GitIndexInventory(inventory.Inventory):
                 else:
                     ie = self.add_path(path, kind, file_id,
                         self.add_parents(path))
-                    data = store[sha].data
+                    data = self.store[sha].data
                     if kind == "symlink":
                         ie.symlink_target = data
                     else:
@@ -341,9 +359,11 @@ class GitIndexInventory(inventory.Inventory):
             else:
                 parent_fid = self.add_parents(dirname)
             ie = self.add_path(dirname, 'directory',
-                    self.mapping.generate_file_id(dirname), parent_fid)
+                    self.fileid_map.lookup_file_id(dirname), parent_fid)
             if ie.file_id in self.basis_inv:
                 ie.revision = self.basis_inv[ie.file_id].revision
             file_id = ie.file_id
+        if type(file_id) != str:
+            raise AssertionError
         return file_id
 
