@@ -1,5 +1,5 @@
 # Copyright (C) 2007 Canonical Ltd
-# Copyright (C) 2008-2009 Jelmer Vernooij <jelmer@samba.org>
+# Copyright (C) 2008-2010 Jelmer Vernooij <jelmer@samba.org>
 # Copyright (C) 2008 John Carr
 #
 # This program is free software; you can redistribute it and/or modify
@@ -22,16 +22,11 @@ import base64
 import stat
 
 from bzrlib import (
+    bencode,
     errors,
     foreign,
-    osutils,
     trace,
-    urlutils,
     )
-try:
-    from bzrlib import bencode
-except ImportError:
-    from bzrlib.util import bencode
 from bzrlib.inventory import (
     ROOT_ID,
     )
@@ -40,10 +35,20 @@ from bzrlib.foreign import (
     VcsMappingRegistry,
     ForeignRevision,
     )
+from bzrlib.revision import (
+    NULL_REVISION,
+    )
 from bzrlib.plugins.git.hg import (
     format_hg_metadata,
     extract_hg_metadata,
     )
+from bzrlib.plugins.git.roundtrip import (
+    extract_bzr_metadata,
+    inject_bzr_metadata,
+    BzrGitRevisionMetadata,
+    deserialize_fileid_map,
+    serialize_fileid_map,
+    )
 
 DEFAULT_FILE_MODE = stat.S_IFREG | 0644
 
@@ -64,7 +69,8 @@ def unescape_file_id(file_id):
             elif file_id[i+1] == 's':
                 ret.append(" ")
             else:
-                raise AssertionError("unknown escape character %s" % file_id[i+1])
+                raise AssertionError("unknown escape character %s" %
+                    file_id[i+1])
             i += 1
         i += 1
     return "".join(ret)
@@ -82,45 +88,31 @@ def warn_escaped(commit, num_escaped):
 
 
 def warn_unusual_mode(commit, path, mode):
-    trace.mutter("Unusual file mode %o for %s in %s. Storing as revision property. ",
-                 mode, path, commit)
-
-
-def squash_revision(target_repo, rev):
-    """Remove characters that can't be stored from a revision, if necessary.
-
-    :param target_repo: Repository in which the revision will be stored
-    :param rev: Revision object, will be modified in-place
-    """
-    if not getattr(target_repo._serializer, "squashes_xml_invalid_characters", True):
-        return
-    from bzrlib.xml_serializer import escape_invalid_chars
-    rev.message, num_escaped = escape_invalid_chars(rev.message)
-    if num_escaped:
-        warn_escaped(rev.foreign_revid, num_escaped)
-    if 'author' in rev.properties:
-        rev.properties['author'], num_escaped = escape_invalid_chars(
-            rev.properties['author'])
-        if num_escaped:
-            warn_escaped(rev.foreign_revid, num_escaped)
-    rev.committer, num_escaped = escape_invalid_chars(rev.committer)
-    if num_escaped:
-        warn_escaped(rev.foreign_revid, num_escaped)
+    trace.mutter("Unusual file mode %o for %s in %s. Storing as revision "
+                 "property. ", mode, path, commit)
 
 
 class BzrGitMapping(foreign.VcsMapping):
     """Class that maps between Git and Bazaar semantics."""
     experimental = False
 
+    BZR_FILE_IDS_FILE = None
+
+    BZR_DUMMY_FILE = None
+
     def __init__(self):
         super(BzrGitMapping, self).__init__(foreign_git)
 
     def __eq__(self, other):
-        return type(self) == type(other) and self.revid_prefix == other.revid_prefix
+        return (type(self) == type(other) and 
+                self.revid_prefix == other.revid_prefix)
 
     @classmethod
     def revision_id_foreign_to_bzr(cls, git_rev_id):
         """Convert a git revision id handle to a Bazaar revision id."""
+        from dulwich.protocol import ZERO_SHA
+        if git_rev_id == ZERO_SHA:
+            return NULL_REVISION
         return "%s:%s" % (cls.revid_prefix, git_rev_id)
 
     @classmethod
@@ -135,29 +127,43 @@ class BzrGitMapping(foreign.VcsMapping):
         # We must just hope they are valid UTF-8..
         if path == "":
             return ROOT_ID
+        if type(path) is unicode:
+            path = path.encode("utf-8")
         return escape_file_id(path)
 
+    def is_control_file(self, path):
+        return path in (self.BZR_FILE_IDS_FILE, self.BZR_DUMMY_FILE)
+
     def parse_file_id(self, file_id):
         if file_id == ROOT_ID:
             return ""
         return unescape_file_id(file_id)
 
+    def revid_as_refname(self, revid):
+        import urllib
+        return "refs/bzr/%s" % urllib.quote(revid)
+
     def import_unusual_file_modes(self, rev, unusual_file_modes):
         if unusual_file_modes:
-            ret = [(name, unusual_file_modes[name]) for name in sorted(unusual_file_modes.keys())]
+            ret = [(path, unusual_file_modes[path])
+                   for path in sorted(unusual_file_modes.keys())]
             rev.properties['file-modes'] = bencode.bencode(ret)
 
     def export_unusual_file_modes(self, rev):
         try:
-            return dict([(self.generate_file_id(path), mode) for (path, mode) in bencode.bdecode(rev.properties['file-modes'].encode("utf-8"))])
+            file_modes = rev.properties['file-modes']
         except KeyError:
             return {}
+        else:
+            return dict([(self.generate_file_id(path), mode) for (path, mode) in bencode.bdecode(file_modes.encode("utf-8"))])
 
-    def _generate_git_svn_metadata(self, rev):
+    def _generate_git_svn_metadata(self, rev, encoding):
         try:
-            return "\ngit-svn-id: %s\n" % rev.properties["git-svn-id"].encode("utf-8")
+            git_svn_id = rev.properties["git-svn-id"]
         except KeyError:
             return ""
+        else:
+            return "\ngit-svn-id: %s\n" % git_svn_id.encode(encoding)
 
     def _generate_hg_message_tail(self, rev):
         extra = {}
@@ -167,9 +173,11 @@ class BzrGitMapping(foreign.VcsMapping):
             if name == 'hg:extra:branch':
                 branch = rev.properties['hg:extra:branch']
             elif name.startswith('hg:extra'):
-                extra[name[len('hg:extra:'):]] = base64.b64decode(rev.properties[name])
+                extra[name[len('hg:extra:'):]] = base64.b64decode(
+                    rev.properties[name])
             elif name == 'hg:renames':
-                renames = bencode.bdecode(base64.b64decode(rev.properties['hg:renames']))
+                renames = bencode.bdecode(base64.b64decode(
+                    rev.properties['hg:renames']))
             # TODO: Export other properties as 'bzr:' extras?
         ret = format_hg_metadata(renames, branch, extra)
         assert isinstance(ret, str)
@@ -194,67 +202,173 @@ class BzrGitMapping(foreign.VcsMapping):
         for name, value in extra.iteritems():
             rev.properties['hg:extra:' + name] = base64.b64encode(value)
         if renames:
-            rev.properties['hg:renames'] = base64.b64encode(bencode.bencode([(new, old) for (old, new) in renames.iteritems()]))
+            rev.properties['hg:renames'] = base64.b64encode(bencode.bencode(
+                [(new, old) for (old, new) in renames.iteritems()]))
         return message
 
-    def _decode_commit_message(self, rev, message):
-        return message.decode("utf-8", "replace")
+    def _extract_bzr_metadata(self, rev, message):
+        (message, metadata) = extract_bzr_metadata(message)
+        return message, metadata
 
-    def _encode_commit_message(self, rev, message):
-        return message.encode("utf-8")
+    def _decode_commit_message(self, rev, message, encoding):
+        return message.decode(encoding), BzrGitRevisionMetadata()
 
-    def export_commit(self, rev, tree_sha, parent_lookup):
+    def _encode_commit_message(self, rev, message, encoding):
+        return message.encode(encoding)
+
+    def export_fileid_map(self, fileid_map):
+        """Export a file id map to a fileid map.
+
+        :param fileid_map: File id map, mapping paths to file ids
+        :return: A Git blob object
+        """
+        from dulwich.objects import Blob
+        b = Blob()
+        b.set_raw_chunks(serialize_fileid_map(fileid_map))
+        return b
+
+    def export_commit(self, rev, tree_sha, parent_lookup, roundtrip):
         """Turn a Bazaar revision in to a Git commit
 
         :param tree_sha: Tree sha for the commit
-        :param parent_lookup: Function for looking up the GIT sha equiv of a bzr revision
+        :param parent_lookup: Function for looking up the GIT sha equiv of a
+            bzr revision
         :return dulwich.objects.Commit represent the revision:
         """
         from dulwich.objects import Commit
         commit = Commit()
         commit.tree = tree_sha
+        if roundtrip:
+            metadata = BzrGitRevisionMetadata()
+        else:
+            metadata = None
+        parents = []
         for p in rev.parent_ids:
-            git_p = parent_lookup(p)
+            try:
+                git_p = parent_lookup(p)
+            except KeyError:
+                git_p = None
+                if metadata is not None:
+                    metadata.explicit_parent_ids = rev.parent_ids
             if git_p is not None:
                 assert len(git_p) == 40, "unexpected length for %r" % git_p
-                commit.parents.append(git_p)
-        commit.committer = fix_person_identifier(rev.committer.encode("utf-8"))
-        commit.author = fix_person_identifier(rev.get_apparent_authors()[0].encode("utf-8"))
+                parents.append(git_p)
+        commit.parents = parents
+        try:
+            encoding = rev.properties['git-explicit-encoding']
+        except KeyError:
+            encoding = rev.properties.get('git-implicit-encoding', 'utf-8')
+        commit.encoding = rev.properties.get('git-explicit-encoding')
+        commit.committer = fix_person_identifier(rev.committer.encode(
+            encoding))
+        commit.author = fix_person_identifier(
+            rev.get_apparent_authors()[0].encode(encoding))
         commit.commit_time = long(rev.timestamp)
         if 'author-timestamp' in rev.properties:
             commit.author_time = long(rev.properties['author-timestamp'])
         else:
             commit.author_time = commit.commit_time
+        commit._commit_timezone_neg_utc = "commit-timezone-neg-utc" in rev.properties
         commit.commit_timezone = rev.timezone
+        commit._author_timezone_neg_utc = "author-timezone-neg-utc" in rev.properties
         if 'author-timezone' in rev.properties:
             commit.author_timezone = int(rev.properties['author-timezone'])
         else:
             commit.author_timezone = commit.commit_timezone
-        commit.message = self._encode_commit_message(rev, rev.message)
+        commit.message = self._encode_commit_message(rev, rev.message, 
+            encoding)
+        assert type(commit.message) == str
+        if metadata is not None:
+            try:
+                mapping_registry.parse_revision_id(rev.revision_id)
+            except errors.InvalidRevisionId:
+                metadata.revision_id = rev.revision_id
+            mapping_properties = set(
+                ['author', 'author-timezone', 'author-timezone-neg-utc',
+                 'commit-timezone-neg-utc', 'git-implicit-encoding',
+                 'git-explicit-encoding', 'author-timestamp', 'file-modes'])
+            for k, v in rev.properties.iteritems():
+                if not k in mapping_properties:
+                    metadata.properties[k] = v
+        if self.roundtripping:
+            commit.message = inject_bzr_metadata(commit.message, metadata, 
+                                                 encoding)
+        assert type(commit.message) == str
         return commit
 
-    def import_commit(self, commit):
+    def import_fileid_map(self, blob):
+        """Convert a git file id map blob.
+
+        :param blob: Git blob object with fileid map
+        :return: Dictionary mapping paths to file ids
+        """
+        return deserialize_fileid_map(blob.data)
+
+    def import_commit(self, commit, lookup_parent_revid):
         """Convert a git commit to a bzr revision.
 
-        :return: a `bzrlib.revision.Revision` object.
+        :return: a `bzrlib.revision.Revision` object and a 
+            dictionary of path -> file ids
         """
         if commit is None:
             raise AssertionError("Commit object can't be None")
-        rev = ForeignRevision(commit.id, self, self.revision_id_foreign_to_bzr(commit.id))
-        rev.parent_ids = tuple([self.revision_id_foreign_to_bzr(p) for p in commit.parents])
-        rev.committer = str(commit.committer).decode("utf-8", "replace")
-        if commit.committer != commit.author:
-            rev.properties['author'] = str(commit.author).decode("utf-8", "replace")
-
+        rev = ForeignRevision(commit.id, self,
+                self.revision_id_foreign_to_bzr(commit.id))
+        rev.parent_ids = tuple([lookup_parent_revid(p) for p in commit.parents])
+        rev.git_metadata = None
+        def decode_using_encoding(rev, commit, encoding):
+            rev.committer = str(commit.committer).decode(encoding)
+            if commit.committer != commit.author:
+                rev.properties['author'] = str(commit.author).decode(encoding)
+            rev.message, rev.git_metadata = self._decode_commit_message(
+                rev, commit.message, encoding)
+        if commit.encoding is not None:
+            rev.properties['git-explicit-encoding'] = commit.encoding
+            decode_using_encoding(rev, commit, commit.encoding)
+        else:
+            for encoding in ('utf-8', 'latin1'):
+                try:
+                    decode_using_encoding(rev, commit, encoding)
+                except UnicodeDecodeError:
+                    pass
+                else:
+                    if encoding != 'utf-8':
+                        rev.properties['git-implicit-encoding'] = encoding
+                    break
         if commit.commit_time != commit.author_time:
             rev.properties['author-timestamp'] = str(commit.author_time)
         if commit.commit_timezone != commit.author_timezone:
-            rev.properties['author-timezone'] = "%d" % (commit.author_timezone, )
+            rev.properties['author-timezone'] = "%d" % commit.author_timezone
+        if commit._author_timezone_neg_utc:
+            rev.properties['author-timezone-neg-utc'] = ""
+        if commit._commit_timezone_neg_utc:
+            rev.properties['commit-timezone-neg-utc'] = ""
         rev.timestamp = commit.commit_time
         rev.timezone = commit.commit_timezone
-        rev.message = self._decode_commit_message(rev, commit.message)
+        if rev.git_metadata is not None:
+            md = rev.git_metadata
+            if md.revision_id:
+                rev.revision_id = md.revision_id
+            if md.explicit_parent_ids:
+                rev.parent_ids = md.explicit_parent_ids
+            rev.properties.update(md.properties)
         return rev
 
+    def get_fileid_map(self, lookup_object, tree_sha):
+        """Obtain a fileid map for a particular tree.
+
+        :param lookup_object: Function for looking up an object
+        :param tree_sha: SHA of the root tree
+        :return: GitFileIdMap instance
+        """
+        try:
+            file_id_map_sha = lookup_object(tree_sha)[self.BZR_FILE_IDS_FILE][1]
+        except KeyError:
+            file_ids = {}
+        else:
+            file_ids = self.import_fileid_map(lookup_object(file_id_map_sha))
+        return GitFileIdMap(file_ids, self)
+
 
 class BzrGitMappingv1(BzrGitMapping):
     revid_prefix = 'git-v1'
@@ -267,28 +381,37 @@ class BzrGitMappingv1(BzrGitMapping):
 class BzrGitMappingExperimental(BzrGitMappingv1):
     revid_prefix = 'git-experimental'
     experimental = True
+    roundtripping = True
 
-    def _decode_commit_message(self, rev, message):
+    BZR_FILE_IDS_FILE = '.bzrfileids'
+
+    BZR_DUMMY_FILE = '.bzrdummy'
+
+    def _decode_commit_message(self, rev, message, encoding):
         message = self._extract_hg_metadata(rev, message)
         message = self._extract_git_svn_metadata(rev, message)
-        return message.decode("utf-8", "replace")
+        message, metadata = self._extract_bzr_metadata(rev, message)
+        return message.decode(encoding), metadata
 
-    def _encode_commit_message(self, rev, message):
-        ret = message.encode("utf-8")
+    def _encode_commit_message(self, rev, message, encoding):
+        ret = message.encode(encoding)
         ret += self._generate_hg_message_tail(rev)
-        ret += self._generate_git_svn_metadata(rev)
+        ret += self._generate_git_svn_metadata(rev, encoding)
         return ret
 
-    def import_commit(self, commit):
-        rev = super(BzrGitMappingExperimental, self).import_commit(commit)
+    def import_commit(self, commit, lookup_parent_revid):
+        rev, file_ids = super(BzrGitMappingExperimental, self).import_commit(commit, lookup_parent_revid)
         rev.properties['converted_revision'] = "git %s\n" % commit.id
-        return rev
+        return rev, file_ids
 
 
 class GitMappingRegistry(VcsMappingRegistry):
     """Registry with available git mappings."""
 
     def revision_id_bzr_to_foreign(self, bzr_revid):
+        if bzr_revid == NULL_REVISION:
+            from dulwich.protocol import ZERO_SHA
+            return ZERO_SHA, None
         if not bzr_revid.startswith("git-"):
             raise errors.InvalidRevisionId(bzr_revid, None)
         (mapping_version, git_sha) = bzr_revid.split(":", 1)
@@ -300,9 +423,9 @@ class GitMappingRegistry(VcsMappingRegistry):
 
 mapping_registry = GitMappingRegistry()
 mapping_registry.register_lazy('git-v1', "bzrlib.plugins.git.mapping",
-                                   "BzrGitMappingv1")
-mapping_registry.register_lazy('git-experimental', "bzrlib.plugins.git.mapping",
-                                   "BzrGitMappingExperimental")
+    "BzrGitMappingv1")
+mapping_registry.register_lazy('git-experimental',
+    "bzrlib.plugins.git.mapping", "BzrGitMappingExperimental")
 mapping_registry.set_default('git-v1')
 
 
@@ -336,18 +459,13 @@ foreign_git = ForeignGit()
 default_mapping = mapping_registry.get_default()()
 
 
-def text_to_blob(texts, entry):
-    from dulwich.objects import Blob
-    text = texts.get_record_stream([(entry.file_id, entry.revision)], 'unordered', True).next().get_bytes_as('fulltext')
-    blob = Blob()
-    blob._text = text
-    return blob
-
-
 def symlink_to_blob(entry):
     from dulwich.objects import Blob
     blob = Blob()
-    blob._text = entry.symlink_target
+    symlink_target = entry.symlink_target
+    if type(symlink_target) == unicode:
+        symlink_target = symlink_target.encode('utf-8')
+    blob.data = symlink_target
     return blob
 
 
@@ -381,7 +499,10 @@ def object_mode(kind, executable):
     if kind == 'directory':
         return stat.S_IFDIR
     elif kind == 'symlink':
-        return stat.S_IFLNK
+        mode = stat.S_IFLNK
+        if executable:
+            mode |= 0111
+        return mode
     elif kind == 'file':
         mode = stat.S_IFREG | 0644
         if executable:
@@ -399,10 +520,18 @@ def entry_mode(entry):
     return object_mode(entry.kind, entry.executable)
 
 
-def directory_to_tree(entry, lookup_ie_sha1, unusual_modes):
-    from dulwich.objects import Tree
+def directory_to_tree(entry, lookup_ie_sha1, unusual_modes, empty_file_name):
+    """Create a Git Tree object from a Bazaar directory.
+
+    :param entry: Inventory entry
+    :param lookup_ie_sha1: Lookup the Git SHA1 for a inventory entry
+    :param unusual_modes: Dictionary with unusual file modes by file ids
+    :param empty_file_name: Name to use for dummy files in empty directories,
+        None to ignore empty directories.
+    """
+    from dulwich.objects import Blob, Tree
     tree = Tree()
-    for name in sorted(entry.children.keys()):
+    for name, value in entry.children.iteritems():
         ie = entry.children[name]
         try:
             mode = unusual_modes[ie.file_id]
@@ -413,76 +542,51 @@ def directory_to_tree(entry, lookup_ie_sha1, unusual_modes):
             tree.add(mode, name.encode("utf-8"), hexsha)
     if entry.parent_id is not None and len(tree) == 0:
         # Only the root can be an empty tree
-        return None
-    tree.serialize()
+        if empty_file_name is not None:
+            tree.add(stat.S_IFREG | 0644, empty_file_name, 
+                Blob().id)
+        else:
+            return None
     return tree
 
 
 def extract_unusual_modes(rev):
     try:
-        foreign_revid, mapping = mapping_registry.parse_revision_id(rev.revision_id)
+        foreign_revid, mapping = mapping_registry.parse_revision_id(
+            rev.revision_id)
     except errors.InvalidRevisionId:
         return {}
     else:
         return mapping.export_unusual_file_modes(rev)
 
 
-def inventory_to_tree_and_blobs(inventory, texts, mapping, unusual_modes, cur=None):
-    """Convert a Bazaar tree to a Git tree.
-
-    :return: Yields tuples with object sha1, object and path
-    """
-    from dulwich.objects import Tree
-    import stat
-    stack = []
-    if cur is None:
-        cur = ""
-    tree = Tree()
-
-    # stack contains the set of trees that we haven't
-    # finished constructing
-    for path, entry in inventory.iter_entries():
-        while stack and not path.startswith(osutils.pathjoin(cur, "")):
-            # We've hit a file that's not a child of the previous path
-            tree.serialize()
-            sha = tree.id
-            yield sha, tree, cur.encode("utf-8")
-            mode = unusual_modes.get(cur.encode("utf-8"), stat.S_IFDIR)
-            t = (mode, urlutils.basename(cur).encode('UTF-8'), sha)
-            cur, tree = stack.pop()
-            tree.add(*t)
-
-        if entry.kind == "directory":
-            stack.append((cur, tree))
-            cur = path
-            tree = Tree()
-        else:
-            if entry.kind == "file":
-                blob = text_to_blob(texts, entry)
-            elif entry.kind == "symlink":
-                blob = symlink_to_blob(entry)
-            else:
-                raise AssertionError("Unknown kind %s" % entry.kind)
-            sha = blob.id
-            yield sha, blob, path.encode("utf-8")
-            name = urlutils.basename(path).encode("utf-8")
-            mode = unusual_modes.get(path.encode("utf-8"), entry_mode(entry))
-            tree.add(mode, name, sha)
-
-    while len(stack) > 1:
-        tree.serialize()
-        sha = tree.id
-        yield sha, tree, cur.encode("utf-8")
-        mode = unusual_modes.get(cur.encode('utf-8'), stat.S_IFDIR)
-        t = (mode, urlutils.basename(cur).encode('UTF-8'), sha)
-        cur, tree = stack.pop()
-        tree.add(*t)
-
-    tree.serialize()
-    yield tree.id, tree, cur.encode("utf-8")
-
-
 def parse_git_svn_id(text):
     (head, uuid) = text.rsplit(" ", 1)
     (full_url, rev) = head.rsplit("@", 1)
     return (full_url, int(rev), uuid)
+
+
+class GitFileIdMap(object):
+
+    def __init__(self, file_ids, mapping):
+        self.file_ids = file_ids
+        self.paths = None
+        self.mapping = mapping
+
+    def lookup_file_id(self, path):
+        try:
+            file_id = self.file_ids[path]
+        except KeyError:
+            file_id = self.mapping.generate_file_id(path)
+        assert type(file_id) is str
+        return file_id
+
+    def lookup_path(self, file_id):
+        if self.paths is None:
+            self.paths = {}
+            for k, v in self.file_ids.iteritems():
+                self.paths[v] = k
+        try:
+            return self.paths[file_id]
+        except KeyError:
+            return self.mapping.parse_file_id(file_id)
