@@ -14,9 +14,9 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-import bzrlib
 from bzrlib import (
     config,
+    debug,
     tag,
     trace,
     ui,
@@ -29,9 +29,6 @@ from bzrlib.errors import (
     NoSuchRevision,
     NotLocalUrl,
     )
-from bzrlib.trace import (
-    info,
-    )
 from bzrlib.transport import (
     Transport,
     )
@@ -43,7 +40,6 @@ lazy_check_versions()
 
 from bzrlib.plugins.git.branch import (
     GitBranch,
-    extract_tags,
     )
 from bzrlib.plugins.git.errors import (
     GitSmartRemoteNotSupported,
@@ -58,6 +54,10 @@ from bzrlib.plugins.git.mapping import (
 from bzrlib.plugins.git.repository import (
     GitRepository,
     )
+from bzrlib.plugins.git.refs import (
+    extract_tags,
+    branch_name_to_ref,
+    )
 
 import dulwich as git
 from dulwich.errors import (
@@ -65,12 +65,13 @@ from dulwich.errors import (
     )
 from dulwich.pack import (
     Pack,
-    PackData,
+    ThinPackData,
     )
 import os
 import tempfile
 import urllib
 import urlparse
+urlparse.uses_netloc.extend(['git', 'git+ssh'])
 
 from dulwich.pack import load_pack_index
 
@@ -81,15 +82,30 @@ def get_test_permutations():
     return []
 
 
+def split_git_url(url):
+    """Split a Git URL.
+
+    :param url: Git URL
+    :return: Tuple with host, port, username, path.
+    """
+    (scheme, netloc, loc, _, _) = urlparse.urlsplit(url)
+    path = urllib.unquote(loc)
+    if path.startswith("/~"):
+        path = path[1:]
+    (username, hostport) = urllib.splituser(netloc)
+    (host, port) = urllib.splitnport(hostport, None)
+    return (host, port, username, path)
+
+
 class GitSmartTransport(Transport):
 
     def __init__(self, url, _client=None):
         Transport.__init__(self, url)
-        (scheme, _, loc, _, _) = urlparse.urlsplit(url)
-        hostport, escaped_path = urllib.splithost(loc)
-        self._path = urllib.unquote(escaped_path)
-        (self._username, hostport) = urllib.splituser(hostport)
-        (self._host, self._port) = urllib.splitnport(hostport, None)
+        (self._host, self._port, self._username, self._path) = \
+            split_git_url(url)
+        if 'transport' in debug.debug_flags:
+            trace.mutter('host: %r, user: %r, port: %r, path: %r',
+                         self._host, self._username, self._port, self._path)
         self._client = _client
 
     def external_url(self):
@@ -107,7 +123,7 @@ class GitSmartTransport(Transport):
     def fetch_pack(self, determine_wants, graph_walker, pack_data, progress=None):
         if progress is None:
             def progress(text):
-                info("git: %s" % text)
+                trace.info("git: %s" % text)
         client = self._get_client(thin_packs=False)
         try:
             return client.fetch_pack(self._get_path(), determine_wants,
@@ -148,8 +164,8 @@ class TCPGitSmartTransport(GitSmartTransport):
             ret = self._client
             self._client = None
             return ret
-        return git.client.TCPGitClient(self._host, self._port, thin_packs=thin_packs,
-            report_activity=self._report_activity)
+        return git.client.TCPGitClient(self._host, self._port,
+            thin_packs=thin_packs, report_activity=self._report_activity)
 
 
 class SSHGitSmartTransport(GitSmartTransport):
@@ -166,8 +182,17 @@ class SSHGitSmartTransport(GitSmartTransport):
             ret = self._client
             self._client = None
             return ret
-        return git.client.SSHGitClient(self._host, self._port, self._username,
+        location_config = config.LocationConfig(self.base)
+        client = git.client.SSHGitClient(self._host, self._port, self._username,
             thin_packs=thin_packs, report_activity=self._report_activity)
+        # Set up alternate pack program paths
+        upload_pack = location_config.get_user_option('git_upload_pack')
+        if upload_pack:
+            client.alternative_paths["upload-pack"] = upload_pack
+        receive_pack = location_config.get_user_option('git_receive_pack')
+        if receive_pack:
+            client.alternative_paths["receive-pack"] = receive_pack
+        return client
 
 
 class RemoteGitDir(GitDir):
@@ -179,13 +204,17 @@ class RemoteGitDir(GitDir):
         self._lockfiles = lockfiles
         self._mode_check_done = None
 
+    def _branch_name_to_ref(self, name, default=None):
+        return branch_name_to_ref(name, default=default)
+
     def open_repository(self):
         return RemoteGitRepository(self, self._lockfiles)
 
-    def open_branch(self, ignore_fallbacks=False):
+    def _open_branch(self, name=None, ignore_fallbacks=False, 
+                    unsupported=False):
         repo = self.open_repository()
-        # TODO: Support for multiple branches in one bzrdir in bzrlib!
-        return RemoteGitBranch(self, repo, "HEAD", self._lockfiles)
+        refname = self._branch_name_to_ref(name)
+        return RemoteGitBranch(self, repo, refname, self._lockfiles)
 
     def open_workingtree(self, recommend_upgrade=False):
         raise NotLocalUrl(self.transport.base)
@@ -206,7 +235,7 @@ class TemporaryPackIterator(Pack):
     @property
     def data(self):
         if self._data is None:
-            self._data = PackData(self._data_path)
+            self._data = ThinPackData(self.resolve_ext_ref, self._data_path)
         return self._data
 
     @property
@@ -217,7 +246,7 @@ class TemporaryPackIterator(Pack):
                 try:
                     def report_progress(cur, total):
                         pb.update("generating index", cur, total)
-                    self.data.create_index(self._idx_path, self.resolve_ext_ref,
+                    self.data.create_index(self._idx_path, 
                         progress=report_progress)
                 finally:
                     pb.finished()
@@ -283,6 +312,15 @@ class RemoteGitRepository(GitRepository):
         except InvalidRevisionId:
             raise NoSuchRevision(self, bzr_revid)
 
+    def lookup_foreign_revision_id(self, foreign_revid, mapping=None):
+        """Lookup a revision id.
+
+        """
+        if mapping is None:
+            mapping = self.get_mapping()
+        # Not really an easy way to parse foreign revids here..
+        return mapping.revision_id_foreign_to_bzr(foreign_revid)
+
 
 class RemoteGitTagDict(tag.BasicTags):
 
@@ -304,7 +342,7 @@ class RemoteGitTagDict(tag.BasicTags):
 class RemoteGitBranch(GitBranch):
 
     def __init__(self, bzrdir, repository, name, lockfiles):
-        self._ref = None
+        self._sha = None
         super(RemoteGitBranch, self).__init__(bzrdir, repository, name,
                 lockfiles)
 
@@ -312,7 +350,7 @@ class RemoteGitBranch(GitBranch):
         raise GitSmartRemoteNotSupported()
 
     def last_revision(self):
-        return self.mapping.revision_id_foreign_to_bzr(self.head)
+        return self.lookup_foreign_revision_id(self.head)
 
     def _get_config(self):
         class EmptyConfig(object):
@@ -324,13 +362,15 @@ class RemoteGitBranch(GitBranch):
 
     @property
     def head(self):
-        if self._ref is not None:
-            return self._ref
+        if self._sha is not None:
+            return self._sha
         heads = self.repository.get_refs()
-        if not self.name in heads:
+        name = self.bzrdir._branch_name_to_ref(self.name, "HEAD")
+        if name in heads:
+            self._sha = heads[name]
+        else:
             raise NoSuchRef(self.name)
-        self._ref = heads[self.name]
-        return self._ref
+        return self._sha
 
     def _synchronize_history(self, destination, revision_id):
         """See Branch._synchronize_history()."""
