@@ -17,12 +17,9 @@
 
 """An adapter between a Git Repository and a Bazaar Branch"""
 
-import bzrlib
 from bzrlib import (
     errors,
-    graph,
     inventory,
-    osutils,
     repository,
     revision,
     revisiontree,
@@ -41,7 +38,6 @@ from bzrlib.plugins.git.mapping import (
     )
 from bzrlib.plugins.git.tree import (
     GitRevisionTree,
-    InterGitRevisionTrees,
     )
 from bzrlib.plugins.git.versionedfiles import (
     GitRevisions,
@@ -49,6 +45,11 @@ from bzrlib.plugins.git.versionedfiles import (
     )
 
 
+from dulwich.objects import (
+    Commit,
+    )
+
+
 class GitRepository(ForeignRepository):
     """An adapter to git repositories for bzr."""
 
@@ -68,12 +69,12 @@ class GitRepository(ForeignRepository):
             repository.InterRepository.register_optimiser(optimiser)
 
     def is_shared(self):
-        return True
+        return False
 
     def supports_rich_root(self):
         return True
 
-    def _warn_if_deprecated(self):
+    def _warn_if_deprecated(self, branch=None):
         # This class isn't deprecated
         pass
 
@@ -90,18 +91,11 @@ class GitRepository(ForeignRepository):
         interrepo = repository.InterRepository.get(source, self)
         return interrepo.dfetch(stop_revision)
 
-    def dfetch_refs(self, source, stop_revision):
-        interrepo = repository.InterRepository.get(source, self)
-        return interrepo.dfetch_refs(stop_revision)
-
 
 class LocalGitRepository(GitRepository):
     """Git repository on the file system."""
 
     def __init__(self, gitdir, lockfiles):
-        # FIXME: This also caches negatives. Need to be more careful
-        # about this once we start writing to git
-        self._parents_provider = graph.CachingParentsProvider(self)
         GitRepository.__init__(self, gitdir, lockfiles)
         self.base = gitdir.root_transport.base
         self._git = gitdir._git
@@ -110,21 +104,20 @@ class LocalGitRepository(GitRepository):
         self.inventories = None
         self.texts = GitTexts(self)
 
-    def all_revision_ids(self):
-        ret = set([revision.NULL_REVISION])
-        heads = self._git.refs.as_dict('refs/heads')
-        if heads == {}:
-            return ret
-        bzr_heads = [self.get_mapping().revision_id_foreign_to_bzr(h) for h in heads.itervalues()]
-        ret = set(bzr_heads)
-        graph = self.get_graph()
-        for rev, parents in graph.iter_ancestry(bzr_heads):
-            ret.add(rev)
-        return ret
+    def _iter_revision_ids(self):
+        for sha in self._git.object_store:
+            o = self._git.object_store[sha]
+            if not isinstance(o, Commit):
+                continue
+            rev = self.get_mapping().import_commit(o,
+                self.lookup_foreign_revision_id)
+            yield o.id, rev.revision_id
 
-    def _make_parents_provider(self):
-        """See Repository._make_parents_provider()."""
-        return self._parents_provider
+    def all_revision_ids(self):
+        ret = set([])
+        for git_sha, revid in self._iter_revision_ids():
+            ret.add(revid)
+        return ret
 
     def get_parent_map(self, revids):
         parent_map = {}
@@ -135,15 +128,10 @@ class LocalGitRepository(GitRepository):
                 continue
             hexsha, mapping = self.lookup_bzr_revision_id(revision_id)
             try:
-                commit = self._git.commit(hexsha)
+                commit = self._git[hexsha]
             except KeyError:
                 continue
-            if commit is None:
-                # Older versions of Dulwich used to return None rather than
-                # raise KeyError.
-                continue
-            else:
-                parent_map[revision_id] = [mapping.revision_id_foreign_to_bzr(p) for p in commit.parents]
+            parent_map[revision_id] = [self.lookup_foreign_revision_id(p, mapping) for p in commit.parents]
         return parent_map
 
     def get_ancestry(self, revision_id, topo_sorted=True):
@@ -162,43 +150,65 @@ class LocalGitRepository(GitRepository):
     def get_signature_text(self, revision_id):
         raise errors.NoSuchRevision(self, revision_id)
 
+    def pack(self, hint=None, clean_obsolete_packs=False):
+        self._git.object_store.pack_loose_objects()
+
     def lookup_foreign_revision_id(self, foreign_revid, mapping=None):
         """Lookup a revision id.
 
-        :param revid: Bazaar revision id.
-        :return: Tuple with git revisionid and mapping.
         """
         if mapping is None:
             mapping = self.get_mapping()
-        return mapping.revision_id_foreign_to_bzr(foreign_revid)
+        from dulwich.protocol import (
+            ZERO_SHA,
+            )
+        if foreign_revid == ZERO_SHA:
+            return revision.NULL_REVISION
+        commit = self._git[foreign_revid]
+        rev = mapping.import_commit(commit, lambda x: None)
+        return rev.revision_id
 
     def has_signature_for_revision_id(self, revision_id):
         return False
 
-    def lookup_bzr_revision_id(self, bzr_revid):
+    def lookup_bzr_revision_id(self, bzr_revid, mapping=None):
         try:
             return mapping_registry.revision_id_bzr_to_foreign(bzr_revid)
         except errors.InvalidRevisionId:
-            raise errors.NoSuchRevision(self, bzr_revid)
+            if mapping is None:
+                mapping = self.get_mapping()
+            try:
+                return self._git.refs[mapping.revid_as_refname(bzr_revid)], mapping
+            except KeyError:
+                # Update refs from Git commit objects
+                # FIXME: Hitting this a lot will be very inefficient...
+                for git_sha, revid in self._iter_revision_ids():
+                    self._git.refs[mapping.revid_as_refname(revid)] = git_sha
+                    if revid == bzr_revid:
+                        return git_sha, mapping
+                raise errors.NoSuchRevision(self, bzr_revid)
 
     def get_revision(self, revision_id):
         git_commit_id, mapping = self.lookup_bzr_revision_id(revision_id)
         try:
-            commit = self._git.commit(git_commit_id)
+            commit = self._git[git_commit_id]
         except KeyError:
             raise errors.NoSuchRevision(self, revision_id)
         # print "fetched revision:", git_commit_id
-        revision = mapping.import_commit(commit)
+        revision = mapping.import_commit(commit,
+            self.lookup_foreign_revision_id)
         assert revision is not None
         return revision
 
     def has_revision(self, revision_id):
         try:
-            self.get_revision(revision_id)
+            git_commit_id, mapping = self.lookup_bzr_revision_id(revision_id)
         except errors.NoSuchRevision:
             return False
-        else:
-            return True
+        return (git_commit_id in self._git)
+
+    def has_revisions(self, revision_ids):
+        return set(filter(self.has_revision, revision_ids))
 
     def get_revisions(self, revids):
         return [self.get_revision(r) for r in revids]
@@ -226,6 +236,19 @@ class LocalGitRepository(GitRepository):
         progress=None):
         return self._git.fetch_objects(determine_wants, graph_walker, progress)
 
+    def _get_versioned_file_checker(self, text_key_references=None,
+                        ancestors=None):
+        return GitVersionedFileChecker(self,
+            text_key_references=text_key_references, ancestors=ancestors)
+    
+
+class GitVersionedFileChecker(repository._VersionedFileChecker):
+
+    file_ids = []
+
+    def _check_file_version_parents(self, texts, progress_bar):
+        return {}, []
+
 
 class GitRepositoryFormat(repository.RepositoryFormat):
     """Git repository format."""
@@ -243,7 +266,9 @@ class GitRepositoryFormat(repository.RepositoryFormat):
         return target_repo_format.rich_root_data
 
     def get_foreign_tests_repository_factory(self):
-        from bzrlib.plugins.git.tests.test_repository import ForeignTestsRepositoryFactory
+        from bzrlib.plugins.git.tests.test_repository import (
+            ForeignTestsRepositoryFactory,
+            )
         return ForeignTestsRepositoryFactory()
 
     def network_name(self):
