@@ -29,6 +29,11 @@ from bzrlib.revisionspec import (
 
 
 def valid_git_sha1(hex):
+    """Check if `hex` is a validly formatted Git SHA1.
+    
+    :param hex: Hex string to validate
+    :return: Boolean
+    """
     import binascii
     try:
         binascii.unhexlify(hex)
@@ -51,8 +56,12 @@ class RevisionSpec_git(RevisionSpec):
         from bzrlib.plugins.git.errors import (
             GitSmartRemoteNotSupported,
             )
+        from bzrlib.plugins.git.mapping import (
+            default_mapping,
+            )
 
-        bzr_revid = branch.mapping.revision_id_foreign_to_bzr(sha1)
+        bzr_revid = getattr(branch.repository, "lookup_foreign_revision_id",
+                              default_mapping.revision_id_foreign_to_bzr)(sha1)
         try:
             if branch.repository.has_revision(bzr_revid):
                 history = self._history(branch, bzr_revid)
@@ -62,7 +71,12 @@ class RevisionSpec_git(RevisionSpec):
         raise InvalidRevisionSpec(self.user_spec, branch)
 
     def _history(self, branch, revid):
-        history = list(branch.repository.iter_reverse_revision_history(revid))
+        branch.lock_read()
+        try:
+            history = list(branch.repository.iter_reverse_revision_history(
+                revid))
+        finally:
+            branch.unlock()
         history.reverse()
         return history
 
@@ -79,6 +93,7 @@ class RevisionSpec_git(RevisionSpec):
 
     def _find_short_git_sha1(self, branch, sha1):
         from bzrlib.plugins.git.mapping import (
+            ForeignGit,
             mapping_registry,
             )
         parse_revid = getattr(branch.repository, "lookup_bzr_revision_id",
@@ -91,6 +106,8 @@ class RevisionSpec_git(RevisionSpec):
                     foreign_revid, mapping = parse_revid(revid)
                 except InvalidRevisionId:
                     continue
+                if not isinstance(mapping.vcs, ForeignGit):
+                    continue
                 if foreign_revid.startswith(sha1):
                     history = self._history(branch, revid)
                     return RevisionInfo.from_revision_id(branch, revid, history)
