@@ -31,6 +31,7 @@ from bzrlib import (
     )
 from bzrlib.branch import (
     Branch,
+    InterBranch,
     )
 from bzrlib.bzrdir import (
     BzrDir,
@@ -55,19 +56,20 @@ class TestGitBranch(tests.TestCaseInTempDir):
 
     def test_open_existing(self):
         GitRepo.init('.')
-
-        thebranch = Branch.open('.')
+        d = BzrDir.open('.')
+        thebranch = d.create_branch()
         self.assertIsInstance(thebranch, branch.GitBranch)
 
     def test_repr(self):
         GitRepo.init('.')
-        thebranch = Branch.open('.')
-        self.assertEquals("LocalGitBranch('file://%s/', 'HEAD')" % self.test_dir, repr(thebranch))
+        d = BzrDir.open('.')
+        thebranch = d.create_branch()
+        self.assertEquals("<LocalGitBranch('file://%s/', 'HEAD')>" % self.test_dir, repr(thebranch))
 
     def test_last_revision_is_null(self):
         GitRepo.init('.')
-
-        thebranch = Branch.open('.')
+        thedir = BzrDir.open('.')
+        thebranch = thedir.create_branch()
         self.assertEqual(revision.NULL_REVISION, thebranch.last_revision())
         self.assertEqual((0, revision.NULL_REVISION),
                          thebranch.last_revision_info())
@@ -81,7 +83,6 @@ class TestGitBranch(tests.TestCaseInTempDir):
     def test_last_revision_is_valid(self):
         self.simple_commit_a()
         head = tests.run_git('rev-parse', 'HEAD').strip()
-
         thebranch = Branch.open('.')
         self.assertEqual(default_mapping.revision_id_foreign_to_bzr(head),
                          thebranch.last_revision())
@@ -114,20 +115,21 @@ class TestGitBranch(tests.TestCaseInTempDir):
         self.assertEquals({"foo": default_mapping.revision_id_foreign_to_bzr(reva)},
                           thebranch.tags.get_tag_dict())
 
-        
+
 
 class TestWithGitBranch(tests.TestCaseWithTransport):
 
     def setUp(self):
         tests.TestCaseWithTransport.setUp(self)
         dulwich.repo.Repo.create(self.test_dir)
-        self.git_branch = Branch.open(self.test_dir)
+        d = BzrDir.open(self.test_dir)
+        self.git_branch = d.create_branch()
 
     def test_get_parent(self):
         self.assertIs(None, self.git_branch.get_parent())
 
     def test_get_stacked_on_url(self):
-        self.assertRaises(errors.UnstackableBranchFormat, 
+        self.assertRaises(errors.UnstackableBranchFormat,
             self.git_branch.get_stacked_on_url)
 
     def test_get_physical_lock_status(self):
@@ -143,6 +145,11 @@ class TestGitBranchFormat(tests.TestCase):
     def test_get_format_description(self):
         self.assertEquals("Git Branch", self.format.get_format_description())
 
+    def test_get_network_name(self):
+        self.assertEquals("git", self.format.network_name())
+
+    def test_supports_tags(self):
+        self.assertTrue(self.format.supports_tags())
 
 
 class BranchTests(tests.TestCaseInTempDir):
@@ -158,6 +165,18 @@ class BranchTests(tests.TestCaseInTempDir):
         os.chdir("..")
         return "d", gitsha
 
+    def make_tworev_branch(self):
+        os.mkdir("d")
+        os.chdir("d")
+        GitRepo.init('.')
+        bb = tests.GitBranchBuilder()
+        bb.set_file("foobar", "foo\nbar\n", False)
+        mark1 = bb.commit("Somebody <somebody@someorg.org>", "mymsg")
+        mark2 = bb.commit("Somebody <somebody@someorg.org>", "mymsg")
+        marks = bb.finish()
+        os.chdir("..")
+        return "d", (marks[mark1], marks[mark2])
+
     def clone_git_branch(self, from_url, to_url):
         from_dir = BzrDir.open(from_url)
         to_dir = from_dir.sprout(to_url)
@@ -181,10 +200,55 @@ class BranchTests(tests.TestCaseInTempDir):
         self.assertEquals({"lala": revid}, newbranch.tags.get_tag_dict())
         self.assertEquals([revid], newbranch.repository.all_revision_ids())
 
+    def test_interbranch_pull(self):
+        path, (gitsha1, gitsha2) = self.make_tworev_branch()
+        oldrepo = Repository.open(path)
+        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
+        newbranch = self.make_branch('g')
+        inter_branch = InterBranch.get(Branch.open(path), newbranch)
+        inter_branch.pull()
+        self.assertEquals(revid2, newbranch.last_revision())
+
+    def test_interbranch_pull_noop(self):
+        path, (gitsha1, gitsha2) = self.make_tworev_branch()
+        oldrepo = Repository.open(path)
+        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
+        newbranch = self.make_branch('g')
+        inter_branch = InterBranch.get(Branch.open(path), newbranch)
+        inter_branch.pull()
+        # This is basically "assertNotRaises"
+        inter_branch.pull()
+        self.assertEquals(revid2, newbranch.last_revision())
+
+    def test_interbranch_pull_stop_revision(self):
+        path, (gitsha1, gitsha2) = self.make_tworev_branch()
+        oldrepo = Repository.open(path)
+        revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
+        newbranch = self.make_branch('g')
+        inter_branch = InterBranch.get(Branch.open(path), newbranch)
+        inter_branch.pull(stop_revision=revid1)
+        self.assertEquals(revid1, newbranch.last_revision())
+
+    def test_interbranch_limited_pull(self):
+        path, (gitsha1, gitsha2) = self.make_tworev_branch()
+        oldrepo = Repository.open(path)
+        revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
+        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
+        newbranch = self.make_branch('g')
+        inter_branch = InterBranch.get(Branch.open(path), newbranch)
+        inter_branch.pull(limit=1)
+        self.assertEquals(revid1, newbranch.last_revision())
+        inter_branch.pull(limit=1)
+        self.assertEquals(revid2, newbranch.last_revision())
+
 
 class ForeignTestsBranchFactory(object):
 
     def make_empty_branch(self, transport):
-        return LocalGitBzrDirFormat().initialize_on_transport(transport).open_branch()
+        d = LocalGitBzrDirFormat().initialize_on_transport(transport)
+        return d.create_branch()
 
     make_branch = make_empty_branch
+
+
+
