@@ -34,6 +34,7 @@ from bzrlib.plugins.git.mapping import (
 
 
 class GitRevisionTree(revisiontree.RevisionTree):
+    """Revision tree implementation based on Git objects."""
 
     def __init__(self, repository, revision_id):
         self._revision_id = revision_id
@@ -46,13 +47,16 @@ class GitRevisionTree(revisiontree.RevisionTree):
         except KeyError, r:
             raise errors.NoSuchRevision(repository, revision_id)
         self.tree = commit.tree
-        self._inventory = GitInventory(self.tree, self.mapping, store, 
-                                       revision_id)
+        fileid_map = self.mapping.get_fileid_map(store.__getitem__, self.tree)
+        self._inventory = GitInventory(self.tree, self.mapping, fileid_map,
+            store, revision_id)
 
     def get_revision_id(self):
+        """See RevisionTree.get_revision_id."""
         return self._revision_id
 
     def get_file_text(self, file_id, path=None):
+        """See RevisionTree.get_file_text."""
         if path is not None:
             entry = self._inventory._get_ie(path)
         else:
@@ -62,35 +66,42 @@ class GitRevisionTree(revisiontree.RevisionTree):
         return entry.object.data
 
 
-def tree_delta_from_git_changes(changes, mapping, specific_file=None, 
-                                require_versioned=False):
+def tree_delta_from_git_changes(changes, mapping,
+        (old_fileid_map, new_fileid_map), specific_file=None,
+        require_versioned=False):
     """Create a TreeDelta from two git trees.
-    
-    source and target are iterators over tuples with: 
+
+    source and target are iterators over tuples with:
         (filename, sha, mode)
     """
     ret = delta.TreeDelta()
     for (oldpath, newpath), (oldmode, newmode), (oldsha, newsha) in changes:
+        if mapping.is_control_file(oldpath):
+            oldpath = None
+        if mapping.is_control_file(newpath):
+            newpath = None
+        if oldpath is None and newpath is None:
+            continue
         if oldpath is None:
-            ret.added.append((newpath, mapping.generate_file_id(newpath), mode_kind(newmode)))
+            ret.added.append((newpath, new_fileid_map.lookup_file_id(newpath.encode("utf-8")), mode_kind(newmode)))
         elif newpath is None:
-            ret.removed.append((oldpath, mapping.generate_file_id(oldpath), mode_kind(oldmode)))
+            ret.removed.append((oldpath, old_fileid_map.lookup_file_id(oldpath.encode("utf-8")), mode_kind(oldmode)))
         elif oldpath != newpath:
-            ret.renamed.append((oldpath, newpath, mapping.generate_file_id(oldpath), mode_kind(newmode), (oldsha != newsha), (oldmode != newmode)))
+            ret.renamed.append((oldpath, newpath, old_fileid_map.lookup_file_id(oldpath.encode("utf-8")), mode_kind(newmode), (oldsha != newsha), (oldmode != newmode)))
         elif mode_kind(oldmode) != mode_kind(newmode):
-            ret.kind_changed.append((newpath, mapping.generate_file_id(newpath), mode_kind(oldmode), mode_kind(newmode)))
+            ret.kind_changed.append((newpath, new_fileid_map.lookup_file_id(newpath.encode("utf-8")), mode_kind(oldmode), mode_kind(newmode)))
         elif oldsha != newsha or oldmode != newmode:
-            ret.modified.append((newpath, mapping.generate_file_id(newpath), mode_kind(newmode), (oldsha != newsha), (oldmode != newmode)))
+            ret.modified.append((newpath, new_fileid_map.lookup_file_id(newpath.encode("utf-8")), mode_kind(newmode), (oldsha != newsha), (oldmode != newmode)))
         else:
-            ret.unchanged.append((newpath, mapping.generate_file_id(newpath), mode_kind(newmode)))
+            ret.unchanged.append((newpath, new_fileid_map.lookup_file_id(newpath.encode("utf-8")), mode_kind(newmode)))
     return ret
 
 
-def changes_from_git_changes(changes, mapping, specific_file=None, 
+def changes_from_git_changes(changes, mapping, specific_file=None,
                                 require_versioned=False):
     """Create a iter_changes-like generator from a git stream.
-    
-    source and target are iterators over tuples with: 
+
+    source and target are iterators over tuples with:
         (filename, sha, mode)
     """
     for (oldpath, newpath), (oldmode, newmode), (oldsha, newsha) in changes:
@@ -127,7 +138,10 @@ def changes_from_git_changes(changes, mapping, specific_file=None,
                 newname = newpath
             else:
                 newparent = mapping.generate_file_id(newparentpath)
-        yield fileid, (oldpath, newpath), (oldsha != newsha), (oldpath is not None, newpath is not None), (oldparent, newparent), (oldname, newname), (oldkind, newkind), (oldexe, newexe)
+        yield (fileid, (oldpath, newpath), (oldsha != newsha),
+             (oldpath is not None, newpath is not None),
+             (oldparent, newparent), (oldname, newname),
+             (oldkind, newkind), (oldexe, newexe))
 
 
 class InterGitRevisionTrees(tree.InterTree):
@@ -139,7 +153,7 @@ class InterGitRevisionTrees(tree.InterTree):
 
     @classmethod
     def is_compatible(cls, source, target):
-        return (isinstance(source, GitRevisionTree) and 
+        return (isinstance(source, GitRevisionTree) and
                 isinstance(target, GitRevisionTree))
 
     def compare(self, want_unchanged=False, specific_files=None,
@@ -149,17 +163,25 @@ class InterGitRevisionTrees(tree.InterTree):
             raise AssertionError
         changes = self.source._repository._git.object_store.tree_changes(
             self.source.tree, self.target.tree, want_unchanged=want_unchanged)
-        return tree_delta_from_git_changes(changes, self.target.mapping, 
+        source_fileid_map = self.source.mapping.get_fileid_map(
+            self.source._repository._git.object_store.__getitem__,
+            self.source.tree)
+        target_fileid_map = self.target.mapping.get_fileid_map(
+            self.target._repository._git.object_store.__getitem__,
+            self.target.tree)
+        return tree_delta_from_git_changes(changes, self.target.mapping,
+            (source_fileid_map, target_fileid_map),
             specific_file=specific_files)
 
     def iter_changes(self, include_unchanged=False, specific_files=None,
-        pb=None, extra_trees=[], require_versioned=True, want_unversioned=False):
+        pb=None, extra_trees=[], require_versioned=True,
+        want_unversioned=False):
         if self.source._repository._git.object_store != self.target._repository._git.object_store:
             raise AssertionError
         changes = self.source._repository._git.object_store.tree_changes(
-            self.source.tree, self.target.tree, 
+            self.source.tree, self.target.tree,
             want_unchanged=include_unchanged)
-        return changes_from_git_changes(changes, self.target.mapping, 
+        return changes_from_git_changes(changes, self.target.mapping,
             specific_file=specific_files)
 
 
