@@ -23,9 +23,13 @@ from dulwich.objects import (
     Tree,
     )
 
+from bzrlib import (
+    annotate,
+    )
+
 from bzrlib.versionedfile import (
     AbsentContentFactory,
-    FulltextContentFactory,
+    ChunkedContentFactory,
     VersionedFiles,
     )
 
@@ -39,6 +43,9 @@ class GitRevisions(VersionedFiles):
     def check(self, progressbar=None):
         return True
 
+    def get_annotator(self):
+        return annotate.Annotator(self)
+
     def iterkeys(self):
         for sha in self.object_store:
             if type(sha) == Commit:
@@ -59,9 +66,9 @@ class GitRevisions(VersionedFiles):
             except KeyError:
                 yield AbsentContentFactory(key)
             else:
-                yield FulltextContentFactory(key, 
+                yield ChunkedContentFactory(key, 
                     tuple([(self.repository.lookup_foreign_revision_id(p, mapping),) for p in commit.parents]), None, 
-                    commit.as_raw_string())
+                    commit.as_raw_chunks())
 
     def get_parent_map(self, keys):
         ret = {}
@@ -84,6 +91,9 @@ class GitTexts(VersionedFiles):
     def check(self, progressbar=None):
         return True
 
+    def get_annotator(self):
+        return annotate.Annotator(self)
+
     def add_mpdiffs(self, records):
         raise NotImplementedError(self.add_mpdiffs)
 
@@ -103,9 +113,9 @@ class GitTexts(VersionedFiles):
                 yield AbsentContentFactory(key)
             else:
                 if isinstance(obj, Tree):
-                    yield FulltextContentFactory(key, None, None, "")
+                    yield ChunkedContentFactory(key, None, None, [])
                 elif isinstance(obj, Blob):
-                    yield FulltextContentFactory(key, None, None, obj.data)
+                    yield ChunkedContentFactory(key, None, None, obj.chunked)
                 else:
                     raise AssertionError("file text resolved to %r" % obj)
 
