@@ -25,18 +25,21 @@ import errno
 from dulwich.objects import (
     Blob,
     )
+from dulwich.protocol import (
+    ZERO_SHA,
+    )
 import os
 import stat
 
 from bzrlib import (
     errors,
     ignores,
+    inventory,
     lockable_files,
     lockdir,
     osutils,
     transport,
     tree,
-    urlutils,
     workingtree,
     )
 from bzrlib.decorators import (
@@ -51,7 +54,9 @@ from bzrlib.plugins.git.tree import (
     changes_from_git_changes,
     tree_delta_from_git_changes,
     )
-
+from bzrlib.plugins.git.mapping import (
+    GitFileIdMap,
+    )
 
 IGNORE_FILENAME = ".gitignore"
 
@@ -59,7 +64,7 @@ IGNORE_FILENAME = ".gitignore"
 class GitWorkingTree(workingtree.WorkingTree):
     """A Git working tree."""
 
-    def __init__(self, bzrdir, repo, branch):
+    def __init__(self, bzrdir, repo, branch, index):
         self.basedir = bzrdir.root_transport.local_abspath('.')
         self.bzrdir = bzrdir
         self.repository = repo
@@ -67,7 +72,7 @@ class GitWorkingTree(workingtree.WorkingTree):
         self._branch = branch
         self._transport = bzrdir.transport
 
-        self.controldir = urlutils.join(self.repository._git._controldir, 'bzr')
+        self.controldir = self.bzrdir.transport.local_abspath('bzr')
 
         try:
             os.makedirs(self.controldir)
@@ -78,7 +83,7 @@ class GitWorkingTree(workingtree.WorkingTree):
         self._control_files = lockable_files.LockableFiles(
             transport.get_transport(self.controldir), 'lock', lockdir.LockDir)
         self._format = GitWorkingTreeFormat()
-        self.index = self.repository._git.open_index()
+        self.index = index
         self.views = self._make_views()
         self._detect_case_handling()
 
@@ -172,9 +177,19 @@ class GitWorkingTree(workingtree.WorkingTree):
 
     def _reset_data(self):
         self._inventory_is_modified = False
-        basis_inv = self.repository.get_inventory(self.mapping.revision_id_foreign_to_bzr(self.repository._git.head()))
-        result = GitIndexInventory(basis_inv, self.mapping, self.index,
-            self.repository._git.object_store)
+        try:
+            head = self.repository._git.head()
+        except KeyError, name:
+            raise errors.NotBranchError("branch %s at %s" % (name, self.repository.base))
+        basis_inv = self.repository.get_inventory(self.branch.lookup_foreign_revision_id(head))
+        store = self.repository._git.object_store
+        if head == ZERO_SHA:
+            fileid_map = {}
+            result = inventory.Inventory(root_id=None)
+        else:
+            fileid_map = self.mapping.get_fileid_map(store.__getitem__,
+                store[head].tree)
+            result = GitIndexInventory(basis_inv, fileid_map, self.index, store)
         self._set_inventory(result, dirty=False)
 
     @needs_read_lock
@@ -227,7 +242,22 @@ class InterIndexGitTree(tree.InterTree):
         changes = self._index.changes_from_tree(
             self.source._repository._git.object_store, self.source.tree, 
             want_unchanged=want_unchanged)
+        source_fileid_map = self.source.mapping.get_fileid_map(
+            self.source._repository._git.object_store.__getitem__,
+            self.source.tree)
+        if self.target.mapping.BZR_FILE_IDS_FILE is not None:
+            try:
+                file_id = self.target.path2id(
+                    self.target.mapping.BZR_FILE_IDS_FILE)
+            except errors.NoSuchId:
+                target_fileid_map = {}
+            else:
+                target_fileid_map = self.import_fileid_map(Blob.from_string(self.target.file_text(file_id)))
+        else:
+            target_fileid_map = {}
+        target_fileid_map = GitFileIdMap(target_fileid_map, self.target.mapping)
         ret = tree_delta_from_git_changes(changes, self.target.mapping, 
+            (source_fileid_map, target_fileid_map),
             specific_file=specific_files, require_versioned=require_versioned)
         if want_unversioned:
             for e in self.target.extras():
